import torch
import copy
import logging
import itertools
import numpy as np

from typing import List, Dict, Tuple, Optional, Union

logger = logging.getLogger(__name__)

try:
    from tokenizations import get_alignments
    from textspan import align_spans, get_original_spans
except ImportError as exp:
    logger.warning(f"Failed to import Python libs `pytokenizations` and/or `textspan`. "
                   f"This may break functions including `respan`, `txt_to_token_span`, `token_to_txt_span`, "
                   f"and `respan_text`.")


def respan(src_tokens: List[str],
           tgt_tokens: List[str],
           src_span: Union[List[tuple], Dict[Tuple[int, int], str]]):
    """
    transfer original spans to target spans
    :param src_tokens: source tokens
    :param tgt_tokens: target tokens
    :param src_span: a list of span tuples. The first element in the tuple
    should be the start index and the second should be the end index
    :return: a list of transferred span tuples.
    """

    s2t, _ = get_alignments(src_tokens, tgt_tokens)

    if isinstance(src_span, list):
        tgt_spans = list()
        for span in src_span:
            start = s2t[span[0]][0]
            if span[1] < len(s2t):
                end = s2t[span[1]][-1]
            else:
                end = s2t[-1][-1]
            if end == start:
                end += 1
            tgt_spans.append((start, end))

    elif isinstance(src_span, dict):
        tgt_spans = dict()
        for span, lb in src_span.items():
            start = s2t[span[0]][0]
            if span[1] < len(s2t):
                end = s2t[span[1]][-1]
            else:
                end = s2t[-1][-1]
            if end == start:
                end += 1
            tgt_spans[(start, end)] = lb

    else:
        raise TypeError("Undefined type for `src_span`")

    return tgt_spans


def respan_text(src_txt: str,
                tgt_txt: str,
                src_span: List[tuple]):
    """
    transfer original spans to target spans
    :param src_txt: source txt
    :param tgt_txt: target txt
    :param src_span: a list of span tuples. The first element in the tuple
    should be the start index and the second should be the end index
    :return: a list of transferred span tuples.
    """
    tgt_spans = align_spans(src_span, src_txt, tgt_txt)
    tgt_spans = [s[0] for s in tgt_spans]

    return tgt_spans


def txt_to_token_span(tokens: List[str],
                      text: str,
                      txt_spans):
    """
    Transfer text-domain spans to token-domain spans
    :param tokens: tokens
    :param text: text
    :param txt_spans: text spans tuples: (start, end, ...)
    :return: a list of transferred span tuples.
    """
    token_indices = get_original_spans(tokens, text)

    try:
        token_indices = [item[0] for item in token_indices]
    except IndexError as err:
        logger.error(f"Encountered token(s) not in original text: {err}")

    if isinstance(txt_spans, list):
        tgt_spans = list()
        for txt_span in txt_spans:
            txt_start = txt_span[0]
            txt_end = txt_span[1]
            start = None
            end = None
            for i, (s, e) in enumerate(token_indices):
                if s <= txt_start < e:
                    start = i
                if s <= txt_end <= e:
                    end = i + 1
                if (start is not None) and (end is not None):
                    break
            assert (start is not None) and (end is not None), ValueError("input spans out of scope")
            tgt_spans.append((start, end))

    elif isinstance(txt_spans, dict):
        tgt_spans = dict()
        for txt_span, v in txt_spans.items():
            txt_start = txt_span[0]
            txt_end = txt_span[1]
            start = None
            end = None
            for i, (s, e) in enumerate(token_indices):
                if s <= txt_start < e:
                    start = i
                if txt_start == e:
                    start = i + 1
                if s <= txt_end <= e:
                    end = i + 1
                if (start is not None) and (end is not None):
                    break
            assert (start is not None) and (end is not None), ValueError("input spans out of scope")
            tgt_spans[(start, end)] = v
    else:
        raise NotImplementedError
    return tgt_spans


def token_to_txt_span(tokens: List[str],
                      text: str,
                      token_spans: Union[List[tuple], dict]) -> Union[List[tuple], dict]:
    """
    Transfer text-domain spans to token-domain spans
    :param tokens: tokens
    :param text: text
    :param token_spans: text spans tuples: (start, end, ...)
    :return: a list of transferred span tuples.
    """
    token_indices = get_original_spans(tokens, text)
    try:
        token_indices = [item[0] for item in token_indices]
    except IndexError as e:
        logger.error(f"Encountered token(s) not in original text: {e}")
    if isinstance(token_spans, dict):
        tgt_spans = dict()
        for token_span, value in token_spans.items():
            txt_start = token_indices[token_span[0]][0]
            txt_end = token_indices[token_span[1] - 1][1]
            tgt_spans[(txt_start, txt_end)] = value
    elif isinstance(token_spans, list):
        tgt_spans = list()
        for tk_span in token_spans:
            txt_start = token_indices[tk_span[0]][0]
            txt_end = token_indices[tk_span[1] - 1][1]
            tgt_spans.append((txt_start, txt_end))
    else:
        raise TypeError(f"Unrecognized type of token_spans: {type(token_spans)}")
    return tgt_spans


def span_to_label(labeled_spans: Dict[Tuple[int, int], str],
                  tokens: List[str]) -> List[str]:
    """
    Convert entity spans to labels

    Parameters
    ----------
    labeled_spans: labeled span dictionary: {(start, end): label}
    tokens: a list of tokens, used to check if the spans are valid.

    Returns
    -------
    a list of string labels
    """
    if labeled_spans:
        assert list(labeled_spans.keys())[-1][1] <= len(tokens), ValueError("label spans out of scope!")

    labels = ['O'] * len(tokens)
    for (start, end), label in labeled_spans.items():
        if type(label) == list or type(label) == tuple:
            lb = label[0][0]
        else:
            lb = label
        labels[start] = 'B-' + lb
        if end - start > 1:
            labels[start + 1: end] = ['I-' + lb] * (end - start - 1)

    return labels


def label_to_span(labels: List[str],
                  scheme: Optional[str] = 'BIO') -> dict:
    """
    convert labels to spans
    :param labels: a list of labels
    :param scheme: labeling scheme, in ['BIO', 'BILOU'].
    :return: labeled spans, a list of tuples (start_idx, end_idx, label)
    """
    assert scheme in ['BIO', 'BILOU'], ValueError("unknown labeling scheme")

    labeled_spans = dict()
    i = 0
    while i < len(labels):
        if labels[i] == 'O':
            i += 1
            continue
        else:
            if scheme == 'BIO':
                if labels[i][0] == 'B':
                    start = i
                    ent = labels[i][2:]
                    i += 1
                    try:
                        while labels[i][0] == 'I':
                            # B-ent followed by I- with different entity; discard the incorrect I- labels
                            if labels[i][2:] != ent:
                                break
                            i += 1
                        end = i
                        labeled_spans[(start, end)] = ent
                    except IndexError:
                        end = i
                        labeled_spans[(start, end)] = ent
                        i += 1
                # this should not happen
                elif labels[i][0] == 'I':
                    i += 1
            elif scheme == 'BILOU':
                if labels[i][0] == 'U':
                    start = i
                    end = i + 1
                    ent = labels[i][2:]
                    labeled_spans[(start, end)] = ent
                    i += 1
                elif labels[i][0] == 'B':
                    start = i
                    ent = labels[i][2:]
                    i += 1
                    try:
                        while labels[i][0] != 'L':
                            i += 1
                        end = i
                        labeled_spans[(start, end)] = ent
                    except IndexError:
                        end = i
                        labeled_spans[(start, end)] = ent
                        break
                    i += 1
                else:
                    i += 1

    return labeled_spans


def span_dict_to_list(span_dict: Dict[Tuple[int], str]):
    """
    convert entity label span dictionaries to span list

    Parameters
    ----------
    span_dict

    Returns
    -------
    span_list
    """
    span_list = list()
    for (s, e), v in span_dict.items():
        span_list.append([s, e, v])
    return span_list


def span_list_to_dict(span_list: List[list]) -> Dict[Tuple[int, int], Union[str, tuple]]:
    """
    convert entity label span list to span dictionaries

    Parameters
    ----------
    span_list

    Returns
    -------
    span_dict
    """
    span_dict = dict()
    for span in span_list:
        span_dict[(span[0], span[1])] = span[2]
    return span_dict


def one_hot(x, n_class=None):
    """
    x : LongTensor of shape (batch size, sequence max_seq_length)
    n_class : integer

    Convert batch of integer letter indices to one-hot vectors of dimension S (# of possible x).
    """

    if n_class is None:
        n_class = np.max(x) + 1
    one_hot_vec = np.zeros([int(np.prod(x.shape)), n_class])
    indices = x.reshape([-1])
    one_hot_vec[np.arange(len(indices)), indices] = 1.0
    one_hot_vec = one_hot_vec.reshape(list(x.shape) + [n_class])
    return one_hot_vec


def probs_to_ids(probs: Union[torch.Tensor, np.ndarray]):
    """
    Convert label probability labels to index

    Parameters
    ----------
    probs: label probabilities

    Returns
    -------
    label indices (shape = one_hot_lbs.shape[:-1])
    """
    if isinstance(probs, torch.Tensor):
        probs = probs.detach().cpu().numpy()
    lb_ids = probs.argmax(axis=-1)
    return lb_ids


def ids_to_lbs(ids: Union[torch.Tensor, np.ndarray], label_types: List[str]):
    if isinstance(ids, torch.Tensor):
        ids = ids.detach().cpu().numpy()
    np_map = np.vectorize(lambda lb: label_types[lb])
    return np_map(ids)


def probs_to_lbs(probs: Union[torch.Tensor, np.ndarray], label_types: List[str]):
    """
    Convert label probability labels to index

    Parameters
    ----------
    probs: label probabilities
    label_types: label types, size = probs.shape[-1]

    Returns
    -------
    labels (shape = one_hot_lbs.shape[:-1])
    """
    np_map = np.vectorize(lambda lb: label_types[lb])
    lb_ids = probs_to_ids(probs)
    return np_map(lb_ids)


def entity_to_bio_labels(entities: List[str]):
    bio_labels = ["O"] + ["%s-%s" % (bi, label) for label in entities for bi in "BI"]
    return bio_labels


def merge_list_of_lists(lists: List[list]):
    merged = list(itertools.chain.from_iterable(lists))
    return merged


def sort_tuples_by_element_idx(tups: List[tuple],
                               idx: Optional[int] = 0,
                               reverse: Optional[bool] = False) -> List[tuple]:
    """
    Function to sort the list of tuples by the second item

    Parameters
    ----------
    tups: a list of tuples to sort
    idx: will sort according to the ith element
    reverse: set True to sort in descending order

    Returns
    -------
    sorted tuples
    """
    tup_ = copy.deepcopy(tups)
    tup_.sort(key=lambda x: x[idx], reverse=reverse)
    return tup_


def merge_overlapped_spans(spans: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """
    Merge overlapped spans

    Parameters
    ----------
    spans: input list of spans

    Returns
    -------
    merged spans
    """

    span_sets = list()
    for span in spans:
        span_set = set(range(span[0], span[1]))
        if not span_sets:
            span_sets.append(span_set)
        elif span_sets[-1] & span_set:
            if span_set - span_sets[-1]:
                span_sets[-1] = span_sets[-1] | span_set
        else:
            span_sets.append(span_set)

    merged_spans = list()
    for span_set in span_sets:
        merged_spans.append((min(span_set), max(span_set) + 1))

    return merged_spans


def rand_argmax(x, **kwargs):
    """
    a random tie-breaking argmax
    https://stackoverflow.com/questions/42071597/numpy-argmax-random-tie-breaking/42071648
    """
    return np.argmax(np.random.random(x.shape) * (x == np.amax(x, **kwargs, keepdims=True)), **kwargs)


def lengths_to_mask(lengths: Union[torch.Tensor, List[int]]):
    """
    Convert sequence lengths to boolean mask.

    Parameters
    ----------
    lengths: sequence lengths; Should be Int/Long Tensor with shape (batch_size, )

    Returns
    -------
    torch.Tensor, Boolean Tensor with shape (batch_size, maximum length).
    The paddings are marked as False.
    """
    if isinstance(lengths, list):
        lengths = torch.tensor(lengths, dtype=torch.long)
    max_length = lengths.max()
    padding_masks = torch.arange(max_length)[None, :] < lengths[:, None]
    return padding_masks



import os
import re
import json
import logging
from typing import Optional, List
from string import printable
from dataclasses import dataclass

import torch
import numpy as np

from .config import BaseNERConfig
from ..data import (
    span_to_label,
    span_list_to_dict,
    entity_to_bio_labels,
    one_hot,
)

logger = logging.getLogger(__name__)


@dataclass
class DataInstance:
    def __setitem__(self, k, v):
        self.__dict__.update(zip(k, v) if type(k) is tuple else [(k, v)])


def feature_lists_to_instance_list(instance_class, **kwargs):
    data_points = list()
    keys = tuple(kwargs.keys())

    for feature_point_list in zip(*tuple(kwargs.values())):
        inst = instance_class()
        inst[keys] = feature_point_list
        data_points.append(inst)

    return data_points


def instance_list_to_feature_lists(instance_list: list, feature_names: Optional[List[str]] = None):
    if not feature_names:
        feature_names = list(instance_list[0].__dict__.keys())

    features_lists = list()
    for name in feature_names:
        features_lists.append([getattr(inst, name) for inst in instance_list])

    return features_lists


class Batch:
    def __init__(self, **kwargs):
        super().__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)

    def to(self, device):
        for k, v in self.__dict__.items():
            try:
                setattr(self, k, v.to(device))
            except AttributeError:
                pass
        return self

    def __len__(self):
        for v in list(self.__dict__.values()):
            if isinstance(v, torch.Tensor):
                return v.size(0)


def load_data_from_json(file_dir: str, config: Optional[BaseNERConfig] = None):
    """
    Load data stored in the current data format.


    Parameters
    ----------
    file_dir: file directory
    config: configuration

    """
    with open(file_dir, 'r', encoding='utf-8') as f:
        data_dict = json.load(f)

    # Load meta if exist
    file_loc = os.path.split(file_dir)[0]
    meta_dir = os.path.join(file_loc, 'meta.json')

    if not os.path.isfile(meta_dir):
        raise FileNotFoundError('Meta file does not exist!')

    with open(meta_dir, 'r', encoding='utf-8') as f:
        meta_dict = json.load(f)

    bio_labels = entity_to_bio_labels(meta_dict['entity_types'])
    label_to_id = {lb: i for i, lb in enumerate(bio_labels)}
    np_map = np.vectorize(lambda lb: label_to_id[lb])

    load_all_sources = config is not None and getattr(config, "load_all_sources", False)
    if 'lf_rec' in meta_dict.keys() and not load_all_sources:
        lf_rec_ids = [meta_dict['lf'].index(lf) for lf in meta_dict['lf_rec']]
    else:
        lf_rec_ids = list(range(meta_dict['num_lf']))

    sentence_list = list()
    lbs_list = list()
    w_lbs_list = list()

    for i in range(len(data_dict)):
        data = data_dict[str(i)]
        # get tokens
        tks = [re.sub("[^{}]+".format(printable), "", tk) for tk in data['data']['text']]
        sent_tks = ['[UNK]' if not tk else tk for tk in tks]
        sentence_list.append(sent_tks)
        # get true labels
        lbs = span_to_label(span_list_to_dict(data['label']), sent_tks)
        lbs_list.append(lbs)
        # get lf annotations (weak labels) in one-hot format tensor
        w_lbs = [span_to_label(span_list_to_dict(data['weak_labels'][lf_idx]), sent_tks) for lf_idx in lf_rec_ids]
        w_lbs = np_map(np.asarray(w_lbs).T)
        w_lbs_one_hot = one_hot(w_lbs, n_class=len(bio_labels))
        w_lbs_list.append(torch.from_numpy(w_lbs_one_hot).to(dtype=torch.float))

    # update config
    if config:
        config.sources = meta_dict['lf_rec'] \
            if 'lf_rec' in meta_dict.keys() and not load_all_sources else meta_dict['lf']
        config.entity_types = meta_dict['entity_types']
        config.bio_label_types = bio_labels
        if 'priors' in meta_dict.keys():
            config.src_priors = meta_dict['priors']
        else:
            config.src_priors = {src: {lb: (0.7, 0.7) for lb in config.entity_types} for src in config.sources}

    if config and getattr(config, 'debug_mode', False):
        return sentence_list[:100], lbs_list[:100], w_lbs_list[:100]
    return sentence_list, lbs_list, w_lbs_list


def load_data_from_pt(file_dir: str, config: Optional[BaseNERConfig] = None):
    """
    Load data that are stored as the previous data format.
    For backward compatibility, should not be used in Wrench

    Parameters
    ----------
    file_dir: file directory
    config: configuration

    """
    data_dict = torch.load(file_dir)

    file_loc, file_name = os.path.split(file_dir)
    meta_dir = os.path.join(file_loc, f'{file_name.split("-")[0]}-metadata.json')

    if not os.path.isfile(meta_dir):
        logger.error('Meta file does not exist!')
        raise FileNotFoundError('Meta file does not exist!')

    with open(meta_dir, 'r', encoding='utf-8') as f:
        meta_dict = json.load(f)

    bio_labels = entity_to_bio_labels(meta_dict['labels'])
    label_to_id = {lb: i for i, lb in enumerate(bio_labels)}

    sentence_list = data_dict['sentences']
    annotation_list = data_dict['annotations']
    label_list = [span_to_label(l, s) for s, l in zip(sentence_list, data_dict['labels'])]

    if 'mapping' in meta_dict.keys():
        annotation_list = convert_conll(annotation_list, meta_dict['labels'], meta_dict['mapping'])

    # meta_dict['sources'] equals "source_to_keep" in the macro file

    load_all_sources = config is not None and getattr(config, "load_all_sources", False)
    sources = list(annotation_list[0].keys()) if load_all_sources else meta_dict['sources']
    weak_label_list = [extract_sequence(
        s, a, sources=sources, label_indices=label_to_id
    ) for s, a in zip(sentence_list, annotation_list)]

    # update config
    if config:
        config.sources = sources
        config.entity_types = meta_dict['labels']
        config.bio_label_types = bio_labels
        if 'mapping' not in meta_dict.keys():
            config.src_priors = meta_dict['priors'] if 'priors' in meta_dict.keys() else None
        else:
            mappings = meta_dict['mapping']
            priors = meta_dict['priors']
            entity_types = meta_dict['labels']

            for src, prs in priors.items():
                entity_dict = {e: list() for e in entity_types}
                for src_ent, pr in prs.items():
                    tgt_ent = mappings.get(src_ent, src_ent)
                    if tgt_ent in entity_types:
                        entity_dict[tgt_ent].append(pr)
                for k, v in entity_dict.items():
                    if v:
                        v = (np.asarray(v)).mean(axis=0)
                        entity_dict[k] = [round(v[0], 2), round(v[1], 2)]
                    else:
                        entity_dict[k] = [1E-2, 1E-2]
                priors[src] = entity_dict
            config.src_priors = priors

    if config and getattr(config, 'debug_mode', False):
        return sentence_list[:100], label_list[:100], weak_label_list[:100]
    return sentence_list, label_list, weak_label_list


def convert_conll(src_annotations, label_types, mappings):
    anno_list = list()
    for annotations in src_annotations:
        anno_dict = dict()
        for src, annos in annotations.items():
            if src not in anno_dict:
                anno_dict[src] = dict()
            for span, values in annos.items():
                anno_dict[src][span] = tuple()
                for lb, conf in values:
                    norm_lb = mappings.get(lb, lb)
                    if norm_lb in label_types:
                        anno_dict[src][span] += ((norm_lb, conf),)
        anno_list.append(anno_dict)
    return anno_list


def extract_sequence(sent,
                     annotations,
                     sources,
                     label_indices):
    """
    Convert the annotations of a spacy document into an array of observations of shape
    (nb_sources, nb_bio_labels)
    """
    sequence = torch.zeros([len(sent), len(sources), len(label_indices)], dtype=torch.float)
    for i, source in enumerate(sources):
        sequence[:, i, 0] = 1.0
        assert source in annotations, logger.error(f"source name {source} is not included in the data")
        for (start, end), vals in annotations[source].items():
            for label, conf in vals:
                if start >= len(sent):
                    logger.warning("Encountered incorrect annotation boundary")
                    continue
                elif end > len(sent):
                    logger.warning("Encountered incorrect annotation boundary")
                    end = len(sent)
                sequence[start:end, i, 0] = 0.0

                sequence[start, i, label_indices["B-%s" % label]] = conf
                if end - start > 1:
                    sequence[start + 1: end, i, label_indices["I-%s" % label]] = conf

    return sequence



from lib2to3.pgen2.token import OP
from optparse import Option
import os
import re
import tqdm
import json
import shutil
import logging
from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)


class TqdmLoggingHandler(logging.Handler):
    """
    Don't let logger print interfere with tqdm progress bar
    """
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


# noinspection PyArgumentList
def set_logging(log_dir: Optional[str] = None):
    """
    setup logging
    Last modified: 07/20/21

    Parameters
    ----------
    log_dir: where to save logging file. Leave None to save no log files

    Returns
    -------

    """
    if log_dir and log_dir != 'null':
        log_dir = os.path.abspath(log_dir)
        if not os.path.isdir(os.path.split(log_dir)[0]):
            os.makedirs(os.path.abspath(os.path.normpath(os.path.split(log_dir)[0])))
        if os.path.isfile(log_dir):
            os.remove(log_dir)
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=logging.INFO,
            handlers=[
                # logging.StreamHandler(sys.stdout),
                logging.FileHandler(log_dir),
                TqdmLoggingHandler(),
            ],
        )
    else:
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            level=logging.INFO,
            handlers=[
                # logging.StreamHandler(sys.stdout),
                TqdmLoggingHandler()
            ],
        )


def logging_args(args):
    """
    Logging model arguments into logs
    Last modified: 08/19/21

    Parameters
    ----------
    args: arguments

    Returns
    -------
    None
    """
    arg_elements = {attr: getattr(args, attr) for attr in dir(args) if not callable(getattr(args, attr))
                    and not attr.startswith("__") and not attr.startswith("_")}
    logger.info(f"Parameters: ({type(args)})")
    for arg_element, value in arg_elements.items():
        logger.info(f"  {arg_element}: {value}")


def prettify_json(text, indent=2, collapse_level=4):
    pattern = r"[\r\n]+ {%d,}" % (indent * collapse_level)
    text = re.sub(pattern, ' ', text)
    text = re.sub(r'([\[({])+ +', r'\g<1>', text)
    text = re.sub(r'[\r\n]+ {%d}([])}])' % (indent * (collapse_level-1)), r'\g<1>', text)
    text = re.sub(r'(\S) +([])}])', r'\g<1>\g<2>', text)
    return text


def remove_dir(directory: str):
    """
    Remove a directory and its subtree folders/files
    """
    dirpath = Path(directory)
    if dirpath.exists() and dirpath.is_dir():
        shutil.rmtree(dirpath)
    return None


def init_dir(directory: str):
    """
    Create the target directory. If the directory exists, remove all subtree folders/files in it.
    """

    remove_dir(directory)
    os.makedirs(os.path.normpath(directory))
    return None


def save_json(obj, path: str, collapse_level: Optional[int] = None):
    """
    Save objective to a json file.
    Create this function so that we don't need to worry about creating parent folders every time

    Parameters
    ----------
    obj: the objective to save
    path: the path to save
    collapse_level: set to any collapse value to prettify output json accordingly

    Returns
    -------
    None
    """
    file_dir = os.path.split(path)[0]
    os.makedirs(file_dir, exist_ok=True)

    json_obj = json.dumps(obj, indent=2, ensure_ascii=False)
    if collapse_level:
        json_obj = prettify_json(json_obj, collapse_level=collapse_level)

    with open(path, 'w', encoding='utf-8') as f:
        f.write(json_obj)

    return None


def prettify_json(text, indent=2, collapse_level=4):
    """
    Make json file more readable by collapsing indent levels higher than `collapse_level`.

    Parameters
    ----------
    text: input json text obj
    indent: the indent value of your json text. Notice that this value needs to be larger than 0
    collapse_level: the level from which the program stops adding new lines

    Usage
    -----
    ```
    my_instance = list()  # user-defined serializable data structure
    json_obj = json.dumps(my_instance, indent=2, ensure_ascii=False)
    json_obj = prettify_json(json_text, indent=2, collapse_level=4)
    with open(path_to_file, 'w', encoding='utf=8') as f:
        f.write(json_text)
    ```
    """
    pattern = r"[\r\n]+ {%d,}" % (indent * collapse_level)
    text = re.sub(pattern, ' ', text)
    text = re.sub(r'([\[({])+ +', r'\g<1>', text)
    text = re.sub(r'[\r\n]+ {%d}([])}])' % (indent * (collapse_level-1)), r'\g<1>', text)
    text = re.sub(r'(\S) +([])}])', r'\g<1>\g<2>', text)
    return text



import logging
from typing import Optional, List
from dataclasses import dataclass

from ..base_model.config import BaseNERConfig

logger = logging.getLogger(__name__)


@dataclass
class CHMMBaseConfig(BaseNERConfig):
    """
    Conditional HMM training configuration
    """
    sources: Optional[List[str]] = None
    src_priors: Optional[dict] = None
    d_emb: Optional[int] = None

    @property
    def d_hidden(self) -> "int":
        """
        Returns the HMM hidden dimension, AKA, the number of bio labels
        """
        return self.n_lbs

    @property
    def d_obs(self) -> "int":
        """
        Returns
        -------
        The observation dimension, equals to the number of bio labels
        """
        return self.n_lbs

    @property
    def n_src(self) -> "int":
        """
        Returns
        -------
        The number of sources
        """
        return len(self.sources) if self.sources is not None else 0

    def save(self, file_dir: str, file_name: Optional[str] = 'chmm-config') -> "CHMMBaseConfig":
        BaseNERConfig.save(self, file_dir, file_name)
        return self

    def load(self, file_dir: str, file_name: Optional[str] = 'chmm-config') -> "CHMMBaseConfig":
        BaseNERConfig.load(self, file_dir, file_name)
        return self




import os
import logging
import numpy as np
from typing import Optional

import torch
from torch.utils.data import DataLoader

from .dataset import CHMMBaseDataset
from .dataset import collate_fn as default_collate_fn
from ..base_model.eval import Metric

logger = logging.getLogger(__name__)

OUT_RECALL = 0.9
OUT_PRECISION = 0.8


class CHMMBaseTrainer:
    def __init__(self,
                 config,
                 collate_fn=None,
                 training_dataset=None,
                 valid_dataset=None,
                 test_dataset=None,
                 pretrain_optimizer=None,
                 optimizer=None):

        self._model = None
        self._config = config
        self._training_dataset = training_dataset
        self._valid_dataset = valid_dataset
        self._test_dataset = test_dataset
        self._collate_fn = collate_fn
        self._pretrain_optimizer = pretrain_optimizer
        self._optimizer = optimizer
        self._init_state_prior = None
        self._init_trans_mat = None
        self._init_emiss_mat = None

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, x):
        logger.warning("Updating DirCHMMTrainer.config")
        self._config = x

    @property
    def model(self):
        return self._model

    def initialize_trainer(self):
        """
        Initialize necessary components for training
        Note: Better not change the order

        Returns
        -------
        the initialized trainer
        """
        self.initialize_matrices()
        self.initialize_model()
        self.initialize_optimizers()
        return self

    def initialize_model(self):
        raise NotImplementedError

    def initialize_matrices(self):
        """
        Initialize <HMM> transition and emission matrices

        Returns
        -------
        self
        """
        assert self._training_dataset and self._valid_dataset
        # inject prior knowledge about transition and emission
        self._init_state_prior = torch.zeros(self._config.d_hidden, device=self._config.device) + 1e-2
        self._init_state_prior[0] += 1 - self._init_state_prior.sum()

        intg_obs = list(map(np.array, self._training_dataset.obs + self._valid_dataset.obs))

        # construct/load initial transition matrix
        dataset_dir = os.path.split(self._config.train_path)[0]
        transmat_path = os.path.join(dataset_dir, "init_transmat.pt")
        if getattr(self._config, "load_init_mat", False):
            if os.path.isfile(transmat_path):
                logger.info("Loading initial transition matrix from disk")
                self._init_trans_mat = torch.load(transmat_path)

                # if the loaded transmat does not have the proper shape, re-calculate it.
                s0_transmat, s1_transmat = self._init_trans_mat.shape
                if not (s0_transmat == s1_transmat == self.config.d_obs):
                    self._init_trans_mat = None

        if self._init_trans_mat is None:
            self._init_trans_mat = torch.tensor(initialise_transmat(
                observations=intg_obs, label_set=self._config.bio_label_types
            )[0], dtype=torch.float)

            if getattr(self._config, "save_init_mat", False):
                logger.info("Saving initial transition matrix")
                torch.save(self._init_trans_mat, transmat_path)

        # construct/load initial emission matrix
        emissmat_path = os.path.join(dataset_dir, "init_emissmat.pt")
        if getattr(self._config, "load_init_mat", False):
            if os.path.isfile(emissmat_path):
                logger.info("Loading initial emission matrix from disk")
                self._init_emiss_mat = torch.load(emissmat_path)

                # if the loaded emissmat does not have the proper shape, re-calculate it.
                s0_emissmat, s1_emissmat, s2_emissmat = self._init_emiss_mat.shape
                if not (s0_emissmat == self.config.n_src) and (s1_emissmat == s2_emissmat == self.config.d_obs):
                    self._init_emiss_mat = None

        if self._init_emiss_mat is None:
            self._init_emiss_mat = torch.tensor(initialise_emissions(
                observations=intg_obs, label_set=self._config.bio_label_types,
                sources=self._config.sources, src_priors=self._config.src_priors
            )[0], dtype=torch.float)

            if getattr(self._config, "save_init_mat", False):
                logger.info("Saving initial emission matrix")
                torch.save(self._init_emiss_mat, emissmat_path)

        return self

    def initialize_optimizers(self, optimizer=None, pretrain_optimizer=None):
        self._optimizer = self.get_optimizer() if optimizer is None else optimizer
        self._pretrain_optimizer = self.get_pretrain_optimizer() if pretrain_optimizer is None else pretrain_optimizer

    def get_dataloader(self, dataset, shuffle=False):
        if dataset is not None:
            dataloader = DataLoader(
                dataset=dataset,
                batch_size=self._config.lm_batch_size,
                collate_fn=self._collate_fn if self._collate_fn is not None else default_collate_fn,
                shuffle=shuffle,
                drop_last=False
            )
            return dataloader
        else:
            logger.error('Dataset is not defined')
            raise ValueError("Dataset is not defined!")

    def pretrain_step(self, data_loader, optimizer, trans_, emiss_):
        raise NotImplementedError

    def training_step(self, data_loader, optimizer):
        raise NotImplementedError

    def train(self):
        raise NotImplementedError

    def valid(self) -> Metric:
        self._model.to(self._config.device)
        valid_metrics = self.evaluate(self._valid_dataset)

        logger.info("Validation results:")
        for k, v in valid_metrics.items():
            logger.info(f"  {k}: {v:.4f}")
        return valid_metrics

    def test(self) -> Metric:
        self._model.to(self._config.device)
        test_metrics = self.evaluate(self._test_dataset)

        logger.info("Test results:")
        for k, v in test_metrics.items():
            logger.info(f"  {k}: {v:.4f}")
        return test_metrics

    def evaluate(self, dataset: CHMMBaseDataset):
        raise NotImplementedError

    def predict(self, dataset: CHMMBaseDataset):
        raise NotImplementedError

    def get_pretrain_optimizer(self):
        raise NotImplementedError

    def get_optimizer(self):
        # ----- initialize optimizer -----
        raise NotImplementedError

    def save(self,
             output_dir: Optional[str] = None,
             save_optimizer: Optional[bool] = False,
             model_name: Optional[str] = 'chmm',
             optimizer_name: Optional[str] = 'chmm-optimizer',
             pretrain_optimizer_name: Optional[str] = 'chmm-pretrain-optimizer'):
        """
        Save model parameters as well as trainer parameters

        Parameters
        ----------
        output_dir: model directory
        save_optimizer: whether to save optimizer
        model_name: model name (suffix free)
        optimizer_name: optimizer name (suffix free)
        pretrain_optimizer_name: pretrain optimizer name (suffix free)

        Returns
        -------
        None
        """
        output_dir = output_dir if output_dir is not None else self._config.output_dir
        logger.info(f"Saving model to {output_dir}")

        model_state_dict = self._model.state_dict()
        torch.save(model_state_dict, os.path.join(output_dir, f'{model_name}.bin'))

        self._config.save(output_dir)

        if save_optimizer:
            logger.info("Saving optimizer and scheduler")
            torch.save(self._optimizer.state_dict(),
                       os.path.join(output_dir, f"{optimizer_name}.bin"))
            torch.save(self._pretrain_optimizer.state_dict(),
                       os.path.join(output_dir, f"{pretrain_optimizer_name}.bin"))

        return None

    def load(self,
             input_dir: Optional[str] = None,
             load_optimizer: Optional[bool] = False,
             model_name: Optional[str] = 'chmm',
             optimizer_name: Optional[str] = 'chmm-optimizer',
             pretrain_optimizer_name: Optional[str] = 'chmm-pretrain-optimizer'):
        """
        Load model parameters.

        Parameters
        ----------
        input_dir: model directory
        load_optimizer: whether load other trainer parameters
        model_name: model name (suffix free)
        optimizer_name: optimizer name (suffix free)
        pretrain_optimizer_name: pretrain optimizer name (suffix free)

        Returns
        -------
        self
        """
        input_dir = input_dir if input_dir is not None else self._config.output_dir
        if self._model is not None:
            logger.warning(f"The original model {type(self._model)} in {type(self)} is not None. "
                           f"It will be overwritten by the loaded model!")
        logger.info(f"Loading model from {input_dir}")
        self.initialize_model()
        self._model.load_state_dict(torch.load(os.path.join(input_dir, f'{model_name}.bin')))
        self._model.to(self.config.device)

        if load_optimizer:
            logger.info("Loading optimizer and scheduler")
            if self._optimizer is None:
                self.initialize_optimizers()
            if os.path.isfile(os.path.join(input_dir, f"{optimizer_name}.bin")):
                self._optimizer.load_state_dict(
                    torch.load(os.path.join(input_dir, f"{optimizer_name}.bin"), map_location=self.config.device)
                )
            else:
                logger.warning("Optimizer file does not exist!")
            if os.path.isfile(os.path.join(input_dir, f"{pretrain_optimizer_name}.bin")):
                self._pretrain_optimizer.load_state_dict(
                    torch.load(os.path.join(input_dir, f"{pretrain_optimizer_name}.bin"))
                )
            else:
                logger.warning("Pretrain optimizer file does not exist!")
        return self

    def save_results(self,
                     output_dir: str,
                     valid_results: Optional[Metric] = None,
                     file_name: Optional[str] = 'results',
                     disable_final_valid: Optional[bool] = False,
                     disable_test: Optional[bool] = False,
                     disable_inter_results: Optional[bool] = False) -> None:
        """
        Save training (validation) results

        Parameters
        ----------
        output_dir: output directory, should be a folder
        valid_results: validation results during the training process
        file_name: file name
        disable_final_valid: disable final validation process (getting validation results of the trained model)
        disable_test: disable test process
        disable_inter_results: do not save inter-results

        Returns
        -------
        None
        """
        if not disable_final_valid:
            logger.info("Getting final validation metrics")
            valid_metrics = self.valid()
        else:
            valid_metrics = None

        if not disable_test:
            logger.info("Getting test metrics.")
            test_metrics = self.test()
        else:
            test_metrics = None

        # write validation and test results
        result_file = os.path.join(output_dir, f'{file_name}.txt')
        logger.info(f"Writing results to {result_file}")
        self.write_result(file_path=result_file,
                          valid_results=valid_results,
                          final_valid_metrics=valid_metrics,
                          test_metrics=test_metrics)

        if not disable_inter_results:
            # save validation inter results
            logger.info(f"Saving inter results")
            inter_result_file = os.path.join(output_dir, f'{file_name}-inter.pt')
            torch.save(valid_results.__dict__, inter_result_file)
        return None

    @staticmethod
    def write_result(file_path: str,
                     valid_results: Optional[Metric] = None,
                     final_valid_metrics: Optional[Metric] = None,
                     test_metrics: Optional[Metric] = None) -> None:
        """
        Support functions for saving training results

        Parameters
        ----------
        file_path: where to save results
        valid_results: validation results during the training process
        final_valid_metrics: validation results of the trained model
        test_metrics

        Returns
        -------

        """
        with open(file_path, 'w') as f:
            if valid_results is not None:
                for i in range(len(valid_results)):
                    f.write(f"[Epoch {i + 1}]\n")
                    for k in ['precision', 'recall', 'f1']:
                        f.write(f"  {k}: {valid_results[k][i]:.4f}")
                    f.write("\n")
            if final_valid_metrics is not None:
                f.write(f"[Best Validation]\n")
                for k in ['precision', 'recall', 'f1']:
                    f.write(f"  {k}: {final_valid_metrics[k]:.4f}")
                f.write("\n")
            if test_metrics is not None:
                f.write(f"[Test]\n")
                for k in ['precision', 'recall', 'f1']:
                    f.write(f"  {k}: {test_metrics[k]:.4f}")
                f.write("\n")
        return None


def initialise_startprob(observations,
                         label_set,
                         src_idx=None):
    """
    calculate initial hidden states (not used in our setup since our sequences all begin from
    [CLS], which corresponds to hidden state "O".
    :param src_idx: source index
    :param label_set: a set of all possible label_set
    :param observations: n_instances X seq_len X n_src X d_obs
    :return: probabilities for the initial hidden states
    """
    n_src = observations[0].shape[1]
    logger.info("Constructing start distribution prior...")

    init_counts = np.zeros((len(label_set),))

    if src_idx is not None:
        for obs in observations:
            init_counts[obs[0, src_idx].argmax()] += 1
    else:
        for obs in observations:
            for z in range(n_src):
                init_counts[obs[0, z].argmax()] += 1

    for i, label in enumerate(label_set):
        if i == 0 or label.startswith("B-"):
            init_counts[i] += 1

    startprob_prior = init_counts + 1
    startprob_ = np.random.dirichlet(init_counts + 1E-10)
    return startprob_, startprob_prior


# TODO: try to use a more reliable source to start the transition and emission
def initialise_transmat(observations,
                        label_set,
                        src_idx=None):
    """
    initialize transition matrix
    :param src_idx: the index of the source of which the transition statistics is computed.
                    If None, use all sources
    :param label_set: a set of all possible label_set
    :param observations: n_instances X seq_len X n_src X d_obs
    :return: initial transition matrix and transition counts
    """

    logger.info("Constructing transition matrix prior...")
    n_src = observations[0].shape[1]
    trans_counts = np.zeros((len(label_set), len(label_set)))

    if src_idx is not None:
        for obs in observations:
            for k in range(0, len(obs) - 1):
                trans_counts[obs[k, src_idx].argmax(), obs[k + 1, src_idx].argmax()] += 1
    else:
        for obs in observations:
            for k in range(0, len(obs) - 1):
                for z in range(n_src):
                    trans_counts[obs[k, z].argmax(), obs[k + 1, z].argmax()] += 1

    # update transition matrix with prior knowledge
    for i, label in enumerate(label_set):
        if label.startswith("B-") or label.startswith("I-"):
            trans_counts[i, label_set.index("I-" + label[2:])] += 1
        elif i == 0 or label.startswith("I-"):
            for j, label2 in enumerate(label_set):
                if j == 0 or label2.startswith("B-"):
                    trans_counts[i, j] += 1

    transmat_prior = trans_counts + 1
    # initialize transition matrix with dirichlet distribution
    transmat_ = np.vstack([np.random.dirichlet(trans_counts2 + 1E-10)
                           for trans_counts2 in trans_counts])
    return transmat_, transmat_prior


def initialise_emissions(observations,
                         label_set,
                         sources,
                         src_priors,
                         strength=1000):
    """
    initialize emission matrices
    :param sources: source names
    :param src_priors: source priors
    :param label_set: a set of all possible label_set
    :param observations: n_instances X seq_len X n_src X d_obs
    :param strength: Don't know what this is for
    :return: initial emission matrices and emission counts?
    """

    logger.info("Constructing emission probabilities...")

    obs_counts = np.zeros((len(sources), len(label_set)), dtype=np.float64)
    # extract the total number of observations for each prior
    for obs in observations:
        obs_counts += obs.sum(axis=0)
    for source_index, source in enumerate(sources):
        # increase p(O)
        obs_counts[source_index, 0] += 1
        # increase the "reasonable" observations
        for pos_index, pos_label in enumerate(label_set[1:]):
            if pos_label[2:] in src_priors[source]:
                obs_counts[source_index, pos_index] += 1
    # construct probability distribution from counts
    obs_probs = obs_counts / (obs_counts.sum(axis=1, keepdims=True) + 1E-3)

    # initialize emission matrix
    matrix = np.zeros((len(sources), len(label_set), len(label_set)))

    for source_index, source in enumerate(sources):
        for pos_index, pos_label in enumerate(label_set):

            # Simple case: set P(O=x|Y=x) to be the recall
            recall = 0
            if pos_index == 0:
                recall = OUT_RECALL
            elif pos_label[2:] in src_priors[source]:
                _, recall = src_priors[source][pos_label[2:]]
            matrix[source_index, pos_index, pos_index] = recall

            for pos_index2, pos_label2 in enumerate(label_set):
                if pos_index2 == pos_index:
                    continue
                elif pos_index2 == 0:
                    precision = OUT_PRECISION
                elif pos_label2[2:] in src_priors[source]:
                    precision, _ = src_priors[source][pos_label2[2:]]
                else:
                    precision = 1.0

                # Otherwise, we set the probability to be inversely proportional to the precision
                # and the (unconditional) probability of the observation
                error_prob = (1 - recall) * (1 - precision) * (0.001 + obs_probs[source_index, pos_index2])

                # We increase the probability for boundary errors (i.e. I-ORG -> B-ORG)
                if pos_index > 0 and pos_index2 > 0 and pos_label[2:] == pos_label2[2:]:
                    error_prob *= 5

                # We increase the probability for errors with same boundary (i.e. I-ORG -> I-GPE)
                if pos_index > 0 and pos_index2 > 0 and pos_label[0] == pos_label2[0]:
                    error_prob *= 2

                matrix[source_index, pos_index, pos_index2] = error_prob

            error_indices = [i for i in range(len(label_set)) if i != pos_index]
            error_sum = matrix[source_index, pos_index, error_indices].sum()
            matrix[source_index, pos_index, error_indices] /= (error_sum / (1 - recall) + 1E-5)

    emission_priors = matrix * strength
    emission_probs = matrix
    return emission_probs, emission_priors



import logging
import numpy as np
from tqdm import tqdm
from typing import List, Optional

import torch
from transformers import AutoTokenizer, AutoModel

from .text import split_overlength_bert_input_sequence

logger = logging.getLogger(__name__)


def build_bert_token_embeddings(tk_seq_list: List[List[str]],
                                model_or_name,
                                tokenizer_or_name,
                                max_seq_length: Optional[int] = 512,
                                device: Optional = 'cpu',
                                prepend_cls_embs: Optional[bool] = False) -> List[torch.Tensor]:
    """
    Build the BERT token embeddings of the input sentences

    Parameters
    ----------
    tk_seq_list: a list of token sequences
    model_or_name: a loaded Huggingface BERT model or its name
    tokenizer_or_name: a loaded Huggingface tokenizer or its name
    max_seq_length: maximum input sequence length that the BERT model allow
    device: device
    prepend_cls_embs: whether add the embeddings corresponding the CLS token back at the beginning of each sequence

    Returns
    -------
    List[torch.Tensor]
    """

    tokenizer = AutoTokenizer.from_pretrained(tokenizer_or_name) if isinstance(tokenizer_or_name, str) \
        else tokenizer_or_name
    model = AutoModel.from_pretrained(model_or_name) if isinstance(model_or_name, str) else model_or_name
    model.to(device)

    split_tk_seq_list = list()
    ori2split_ids_map = list()
    n = 0

    # update input sentences so that every sentence has BERT length < 510
    logger.info(f'Checking lengths. Paragraphs longer than {max_seq_length} tokens will be separated.')
    for tk_seq in tk_seq_list:
        tk_seqs = split_overlength_bert_input_sequence(tk_seq, tokenizer, max_seq_length)
        n_splits = len(tk_seqs)
        split_tk_seq_list += tk_seqs

        ori2split_ids_map.append(list(range(n, n + n_splits)))
        n += n_splits

    logger.info('Building embeddings...')
    sent_emb_list = build_emb_helper(split_tk_seq_list, tokenizer, model, device, prepend_cls_embs)

    # Combine embeddings so that the embedding lengths equal to the lengths of the original sentences
    logger.info('Combining results...')
    tk_emb_seq_list = list()

    for sep_ids in ori2split_ids_map:
        cat_emb = None

        for sep_idx in sep_ids:
            if cat_emb is None:
                cat_emb = sent_emb_list[sep_idx]
            else:
                cat_emb = torch.cat([cat_emb, sent_emb_list[sep_idx][1:]], dim=0)

        assert cat_emb is not None, ValueError('Empty embedding!')
        tk_emb_seq_list.append(cat_emb)

    # The embeddings of [CLS] + original tokens
    for emb, tk_seq in zip(tk_emb_seq_list, tk_seq_list):
        if prepend_cls_embs:
            assert len(emb) == len(tk_seq) + 1  # check the length of embeddings equals to the length of sequence
        else:
            assert len(emb) == len(tk_seq)

    return tk_emb_seq_list


def build_emb_helper(tk_seq_list: List[List[str]],
                     tokenizer,
                     model,
                     device: Optional = 'cpu',
                     prepend_cls_embs: Optional[bool] = False):

    tk_emb_seq_list = list()

    for tk_seq in tqdm(tk_seq_list):

        encs = tokenizer(tk_seq, is_split_into_words=True, add_special_tokens=True, return_offsets_mapping=True)
        input_ids = torch.tensor([encs.input_ids], device=device)
        offsets_mapping = np.array(encs.offset_mapping)

        # calculate BERT last layer embeddings
        with torch.no_grad():
            # get the last hidden state from the BERT model
            last_hidden_states = model(input_ids)[0].squeeze(0).to('cpu')
            # remove the token embeddings regarding the [CLS] and [SEP]
            trunc_hidden_states = last_hidden_states[1:-1, :]

        ori2bert_tk_ids = list()
        idx = 0
        for tk_start in (offsets_mapping[1:-1, 0] == 0):
            if tk_start:
                ori2bert_tk_ids.append([idx])
            else:
                ori2bert_tk_ids[-1].append(idx)
            idx += 1

        emb_list = list()
        for ids in ori2bert_tk_ids:
            embeddings = trunc_hidden_states[ids, :]  # first dim could be 1 or n
            emb_list.append(embeddings.mean(dim=0))

        if prepend_cls_embs:
            # add back the embedding of [CLS] as the sentence embedding
            emb_list = [last_hidden_states[0, :]] + emb_list

        bert_emb = torch.stack(emb_list)
        assert not bert_emb.isnan().any(), ValueError('NaN Embeddings!')
        tk_emb_seq_list.append(bert_emb.detach().cpu())

    return tk_emb_seq_list



import os
import logging
import numpy as np
from typing import Optional, List, Any, Dict

import torch
from torch.utils.data import DataLoader

import matplotlib.pyplot as plt

from .model_output import ModelOutput

logger = logging.getLogger(__name__)


def default_collate_fn(features: List[Any]) -> Dict[str, Any]:
    """
    default collate function for pytorch. Copied from `transformers`

    Parameters
    ----------
    features: input features

    Returns
    -------
    Dictionary of feature name: feature value
    """
    if not isinstance(features[0], dict):
        features = [vars(f) for f in features]
    first = features[0]
    batch = {}

    # Special handling for labels.
    # Ensure that tensor is created with the correct type
    # (it should be automatically the case, but let's make sure of it.)
    if "label" in first and first["label"] is not None:
        label = first["label"].item() if isinstance(first["label"], torch.Tensor) else first["label"]
        dtype = torch.long if isinstance(label, int) else torch.float
        batch["labels"] = torch.tensor(np.array([f["label"] for f in features]), dtype=dtype)
    elif "label_ids" in first and first["label_ids"] is not None:
        if isinstance(first["label_ids"], torch.Tensor):
            batch["labels"] = torch.stack([f["label_ids"] for f in features])
        else:
            dtype = torch.long if type(first["label_ids"][0]) is int else torch.float
            batch["labels"] = torch.tensor(np.array([f["label_ids"] for f in features]), dtype=dtype)

    # Handling of all other possible keys.
    # Again, we will use the first element to figure out which key/values are not None for this model.
    for k, v in first.items():
        dtype = torch.long if isinstance(v, int) else torch.float
        if k not in ("label", "label_ids") and v is not None and not isinstance(v, str):
            if isinstance(v, torch.Tensor):
                batch[k] = torch.stack([f[k] for f in features]).to(dtype)
            else:
                batch[k] = torch.tensor(np.array([f[k] for f in features])).to(dtype)

    return batch


class BaseTrainer:
    def __init__(self,
                 config,
                 training_dataset=None,
                 valid_dataset=None,
                 test_dataset=None,
                 collate_fn=default_collate_fn):

        self._config = config
        self._training_dataset = training_dataset
        self._valid_dataset = valid_dataset
        self._test_dataset = test_dataset
        self._collate_fn = collate_fn
        self._model = None
        self._optimizer = None
        self._scheduler = None
        self._device = getattr(config, "device", "cpu")

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, x):
        logger.warning("Updating DirCHMMTrainer.config")
        self._config = x

    @property
    def model(self):
        return self._model

    def initialize(self):
        self.initialize_model()
        self.initialize_optimizer()
        self.initialize_scheduler()
        return self

    def initialize_model(self):
        logger.warning("Model is not defined by initialization function!")
        return self

    def initialize_optimizer(self):
        logger.warning("Optimizer is not defined by initialization function!")
        return self

    def initialize_scheduler(self):
        logger.warning("Scheduler is not defined by initialization function!")
        return self

    def get_dataloader(self,
                       dataset,
                       shuffle: Optional[bool] = False,
                       batch_size: Optional[int] = 0):
        try:
            dataloader = DataLoader(
                dataset=dataset,
                collate_fn=self._collate_fn,
                batch_size=batch_size if batch_size else self._config.batch_size,
                num_workers=getattr(self._config, "num_workers", 0),
                pin_memory=getattr(self._config, "pin_memory", False),
                shuffle=shuffle,
                drop_last=False
            )
        except Exception as e:
            logger.exception(e)
            raise e

        return dataloader

    def run(self, *args, **kwargs):
        """
        The main process of the trainer class
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def training_step(self, dataloader):
        raise NotImplementedError

    def predict(self, dataset):
        """
        Predict labels/values given input dataset

        Parameters
        ----------
        dataset: input dataset

        Returns
        -------
        model output
        """
        data_loader = self.get_dataloader(dataset)

        self._model.to(self._device)
        self._model.eval()

        preds = ModelOutput()

        with torch.no_grad():
            for batch in data_loader:
                # get data and move to device
                for k, v in batch.items():
                    if isinstance(v, torch.Tensor):
                        batch[k] = v.to(self._device)

                # get model output
                preds_batch = self._model(**batch)
                preds += preds_batch

        return preds.unhook()

    def evaluate(self, dataset):
        raise NotImplementedError

    def valid(self):
        metrics = self.evaluate(self._valid_dataset)
        return metrics

    def test(self):
        metrics = self.evaluate(self._test_dataset)
        return metrics

    def save(self,
             output_dir: Optional[str] = None,
             save_optimizer: Optional[bool] = False,
             save_scheduler: Optional[bool] = False,
             model_name: Optional[str] = 'model',
             optimizer_name: Optional[str] = 'optimizer',
             scheduler_name: Optional[str] = 'scheduler'):
        """
        Save model parameters as well as trainer parameters

        Parameters
        ----------
        output_dir: model directory
        save_optimizer: whether to save optimizer
        save_scheduler: whether to save scheduler
        model_name: model name (suffix free)
        optimizer_name: optimizer name (suffix free)
        scheduler_name: scheduler name (suffix free)

        Returns
        -------
        None
        """
        output_dir = output_dir if output_dir is not None else getattr(self._config, 'output_dir', 'output')

        model_state_dict = self._model.state_dict()
        torch.save(model_state_dict, os.path.join(output_dir, f'{model_name}.bin'))

        self._config.save(output_dir)

        if save_optimizer:
            torch.save(self._optimizer.state_dict(), os.path.join(output_dir, f"{optimizer_name}.bin"))
        if save_scheduler and self._scheduler is not None:
            torch.save(self._scheduler.state_dict(), os.path.join(output_dir, f"{scheduler_name}.bin"))

        return None

    def load(self,
             input_dir: Optional[str] = None,
             load_optimizer: Optional[bool] = False,
             load_scheduler: Optional[bool] = False,
             model_name: Optional[str] = 'model',
             optimizer_name: Optional[str] = 'optimizer',
             scheduler_name: Optional[str] = 'scheduler'):
        """
        Load model parameters.

        Parameters
        ----------
        input_dir: model directory
        load_optimizer: whether load other trainer parameters
        load_scheduler: whether load scheduler
        model_name: model name (suffix free)
        optimizer_name: optimizer name (suffix free)
        scheduler_name: scheduler name

        Returns
        -------
        self
        """
        input_dir = input_dir if input_dir is not None else getattr(self._config, 'output_dir', 'output')

        logger.info(f"Loading model from {input_dir}")

        self.initialize_model()
        self._model.load_state_dict(torch.load(os.path.join(input_dir, f'{model_name}.bin')))
        self._model.to(self._device)

        if load_optimizer:
            logger.info("Loading optimizer")

            if self._optimizer is None:
                self.initialize_optimizer()

            if os.path.isfile(os.path.join(input_dir, f"{optimizer_name}.bin")):
                self._optimizer.load_state_dict(
                    torch.load(os.path.join(input_dir, f"{optimizer_name}.bin"), map_location=self._device)
                )
            else:
                logger.warning("Optimizer file does not exist!")

        if load_scheduler:
            logger.info("Loading scheduler")

            if self._scheduler is None:
                self.initialize_scheduler()

            if os.path.isfile(os.path.join(input_dir, f"{scheduler_name}.bin")):
                self._optimizer.load_state_dict(
                    torch.load(os.path.join(input_dir, f"{scheduler_name}.bin"), map_location=self._device)
                )
            else:
                logger.warning("Scheduler file does not exist!")
        return self

    def save_fig(self, fig, epoch: Optional[int] = None, directory="Plots"):
        save_dir = os.path.join(self._config.output_dir, directory)
        os.makedirs(save_dir, exist_ok=True)

        f_name = f'epoch-{epoch}.png' if epoch is not None else 'test'

        fig.savefig(os.path.join(save_dir, f_name), bbox_inches='tight')
        plt.close(plt.gcf())

    @staticmethod
    def log_tensorboard_fig(tensorboard_writer, fig, epoch: Optional[int] = 0, directory="Plots"):
        # Draw figure on canvas
        fig.canvas.draw()

        # Convert the figure to numpy array, read the pixel values and reshape the array
        img = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
        img = img.reshape(fig.canvas.get_width_height()[::-1] + (3,))

        # Normalize into 0-1 range for TensorBoard(X).
        # Swap axes for newer versions where API expects colors in first dim
        img = img / 255.0
        img = np.rot90(np.flipud(img), k=-1)
        img = np.swapaxes(img, 0, 2)  # if your TensorFlow + TensorBoard version are >= 1.8

        # Add figure in numpy "image" to TensorBoard writer
        tensorboard_writer.add_image(directory, img, epoch)
        plt.close(fig)



import os
import logging
import numpy as np
from tqdm.auto import tqdm
from typing import List, Optional, Union

import torch
from torch.utils.data import DataLoader
from transformers import (
    AutoTokenizer,
    BatchEncoding
)

from ..text import split_overlength_bert_input_sequence_legacy
from ..base_model.dataset import load_data_from_json, load_data_from_pt
from ..base_model.config import BaseNERConfig

logger = logging.getLogger(__name__)


class BertNERBaseDataset(torch.utils.data.Dataset):
    def __init__(self,
                 text: Optional[List[List[str]]] = None,
                 lbs: Optional[List[List[str]]] = None,
                 encoded_texts: Optional[BatchEncoding] = BatchEncoding(),
                 encoded_lbs: Optional[List[List[int]]] = None,
                 mapping_ids: Optional[List[List[int]]] = None,
                 token_masks: Optional[List[List[int]]] = None,
                 ):
        super().__init__()
        self._text = text
        self._lbs = lbs
        # split text so that every sentence is within maximum length when they are converted to BERT tokens
        self._encoded_texts = encoded_texts
        self._encoded_lbs = encoded_lbs if encoded_lbs is not None else list()
        # mapping from original sentences to splitted ones ([[1.1, 1.2], [2], [3]])
        self._mapping_ids = mapping_ids if mapping_ids is not None else list()
        # mask out sub-tokens and paddings
        self._token_masks = token_masks if token_masks is not None else list()

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text_: List[List[str]]):
        logger.warning("Setting text instances. Need to run `encode_text` or `encode_text_and_lbs` "
                       "to update encoded text.")
        self._text = text_

    @property
    def lbs(self):
        return self._lbs

    @lbs.setter
    def lbs(self, labels: List[List[str]]):
        assert len(self._text) == len(labels), ValueError("The number of text & labels instances does not match!")
        for txt, lbs_ in zip(self._text, labels):
            assert len(txt) == len(lbs_), ValueError("The lengths of text & labels instances does not match!")
        logger.warning("Setting label instances. Need to run `encode_text_and_lbs` to update encoded labels.")
        self._lbs = labels

    @property
    def token_masks(self):
        return np.asarray(self._token_masks)

    @property
    def encoded_lbs(self):
        return np.asarray(self._encoded_lbs)

    @property
    def mapping_ids(self):
        return self._mapping_ids

    @property
    def n_insts(self):
        return len(self._encoded_texts.input_ids)

    def __len__(self):
        return self.n_insts

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self._encoded_texts.items() if key != 'offset_mapping'}
        item['token_masks'] = torch.tensor(self._token_masks[idx])
        if self._encoded_lbs:
            item['labels'] = torch.tensor(self._encoded_lbs[idx])
        return item

    def encode_text(self, config):
        """
        Encode tokens so that they match the BERT data format

        Parameters
        ----------
        config: configuration file

        Returns
        -------
        self (BertNERDataset)
        """
        assert self._text, ValueError("Need to specify text")
        logger.info("Encoding BERT text")

        tokenizer = AutoTokenizer.from_pretrained(config.bert_model_name_or_path)

        sp_text = list()
        mapping_ids = list()
        for text in self._text:
            # separate a sentence into several pieces if it exceeds the maximum length
            st, st_lens, st_ids = split_overlength_bert_input_sequence_legacy(text, tokenizer, config.max_length)
            sp_text += st

            if not mapping_ids:
                mapping_ids.append(st_ids)
            else:
                mapping_ids.append(st_ids + mapping_ids[-1][-1] + 1)

        self._mapping_ids = mapping_ids

        logger.info('Encoding sentences into BERT tokens')
        self._encoded_texts = tokenizer(self._text,
                                        is_split_into_words=True,
                                        return_offsets_mapping=True,
                                        padding='max_length',
                                        max_length=config.max_length,
                                        truncation=True)

        token_masks = list()
        for doc_offset in tqdm(self._encoded_texts.offset_mapping):
            arr_offset = np.array(doc_offset)

            # create an empty array of False
            masks = np.zeros(len(doc_offset), dtype=bool)
            masks[(arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)] = True
            token_masks.append(masks)
        self._token_masks = token_masks

        return self

    def encode_text_and_lbs(self, config):
        """
        Encode tokens and labels so that they match the BERT data format

        Parameters
        ----------
        config: configuration file

        Returns
        -------
        self (BertNERDataset)
        """
        logger.info("Encoding BERT text and labels")

        assert self._text and self._lbs, ValueError("Need to specify text and labels")

        tokenizer = AutoTokenizer.from_pretrained(config.bert_model_name_or_path)

        sp_text = list()
        sp_lbs = list()
        mapping_ids = list()
        for text, lbs in zip(self._text, self._lbs):
            # break a sentence into several pieces if it exceeds the maximum length
            st, st_lens, st_ids = split_overlength_bert_input_sequence_legacy(text, tokenizer, config.max_length)
            sp_text += st

            start_idx = 0
            for st_len in st_lens:
                sp_lbs.append(lbs[start_idx: st_len])
                start_idx += st_len

            if not mapping_ids:
                mapping_ids.append(st_ids)
            else:
                mapping_ids.append(st_ids + mapping_ids[-1][-1] + 1)

        self._mapping_ids = mapping_ids

        logger.info('Encoding sentences into BERT tokens')
        self._encoded_texts = tokenizer(sp_text,
                                        is_split_into_words=True,
                                        return_offsets_mapping=True,
                                        padding='max_length',
                                        max_length=config.max_length,
                                        truncation=True)

        soft_label_flag = False
        if isinstance(sp_lbs[0], (np.ndarray, torch.Tensor)):
            soft_label_flag = True

        if soft_label_flag:
            labels = sp_lbs
        else:
            labels = [[config.lb2idx[lb] for lb in lbs] for lbs in sp_lbs]

        encoded_labels = list()
        token_masks = list()

        logger.info('Aligning labels to encoded text')
        for doc_labels, doc_offset in tqdm(zip(labels, self._encoded_texts.offset_mapping), total=len(labels)):
            # create an empty array of -100
            if soft_label_flag:
                doc_enc_labels = np.ones([len(doc_offset), config.n_lbs]) * -100
            else:
                doc_enc_labels = np.ones(len(doc_offset), dtype=int) * -100
            arr_offset = np.array(doc_offset)

            # set labels whose first offset position is 0 and the second is not 0
            doc_enc_labels[(arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)] = doc_labels
            encoded_labels.append(doc_enc_labels.tolist())

            # create an empty array of False
            masks = np.zeros(len(doc_offset), dtype=bool)
            masks[(arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)] = True
            token_masks.append(masks)
        self._encoded_lbs = encoded_labels
        self._token_masks = token_masks

        return self

    def select(self, ids: Union[List[int], np.ndarray, torch.Tensor]):
        """
        Select a subset of dataset

        Parameters
        ----------
        ids: instance indices to select

        Returns
        -------
        A BertClassificationDataset consists of selected items
        """
        if np.max(ids) >= self.n_insts:
            logger.error("Invalid indices: exceeding the dataset size!")
            raise ValueError('Invalid indices: exceeding the dataset size!')
        text_ = np.asarray(self._text, dtype=object)[ids].tolist()
        lbs_ = np.asarray(self._lbs, dtype=object)[ids].tolist() if self._lbs else None
        logger.warning("Need to run `encode_text` or `encode_text_and_lbs` on the selected subset.")
        return BertNERBaseDataset(text_, lbs_)

    def load_file(self,
                  file_path: str,
                  config: Optional[BaseNERConfig] = None) -> "BertNERBaseDataset":
        """
        Load data from disk

        Parameters
        ----------
        file_path: the directory of the file. In JSON or PT
        config: chmm configuration; Optional to make function testing easier.

        Returns
        -------
        self (BERTNERDataset)
        """

        file_path = os.path.normpath(file_path)
        logger.info(f'Loading data from {file_path}')

        if file_path.endswith('.json'):
            sentence_list, label_list, weak_label_list = load_data_from_json(file_path, config)
        # for backward compatibility
        elif file_path.endswith('.pt'):
            sentence_list, label_list, weak_label_list = load_data_from_pt(file_path, config)
        else:
            logger.error(f"Unsupported data type: {file_path}")
            raise TypeError(f"Unsupported data type: {file_path}")

        self._text = sentence_list
        self._lbs = label_list
        logger.info(f'Data loaded from {file_path}.')

        return self



import copy
import torch
import numpy as np
from functools import cached_property
from typing import List, Optional, Union, Any


class ModelOutput:
    """
    Sequence labeling metrics

    This class is designed to facilitate easy inheritance.
    You can conveniently add new metrics into the class by reloading the `__init__` function.
    Other member functions should work fine with new metrics.

    all class members should be torch.Tensor
    """
    def __init__(self):
        pass

    def keys(self, idx: Optional[int] = None) -> Union[List[str], str]:
        if idx is None:
            return list(self.__dict__.keys())
        elif isinstance(idx, int):
            return list(self.__dict__.keys())[idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    def items(self, idx: Optional[int] = None):
        if idx is None:
            for k in self.keys():
                yield k, self[k]
        elif isinstance(idx, int):
            for k in self.keys():
                yield k, self[k][idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    def values(self, idx: Optional[int] = None):
        if idx is None:
            for k in self.keys():
                yield self[k]
        elif isinstance(idx, int):
            for k in self.keys():
                yield self[k][idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    @cached_property
    def array_attribs(self):
        """
        Get all tensor members
        """
        ks = self.keys()
        array_ks = list()
        for k in ks:
            if isinstance(getattr(self, k), (torch.Tensor, np.ndarray)):
                array_ks.append(k)
        return array_ks

    @property
    def is_empty(self):
        """
        Check if the instance has any tensor member
        """
        return False if self.array_attribs else True

    def unhook(self):
        """
        Detach the results from training graph and move in to cpu
        """
        for attrib in self.array_attribs:
            detached = getattr(self, attrib).detach().clone().cpu()
            setattr(self, attrib, detached)
        return self

    def numpy(self):
        """
        Convert the Tensor results to numpy.array

        Returns
        -------
        numpy array
        """
        for attrib in self.array_attribs:
            try:
                detached = getattr(self, attrib).numpy()
                setattr(self, attrib, detached)
            except:
                pass
        return self

    def to_dict(self):
        """
        Convert class to dictionary
        """
        return {attrib: getattr(self, attrib) for attrib in self.array_attribs}

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return str(self.to_dict())

    def __len__(self):
        return len(self.array_attribs)

    def __iter__(self):
        for k in self.array_attribs:
            yield self[k]

    def __getitem__(self, item: Union[int, str]):
        if isinstance(item, str):
            return getattr(self, item)
        elif isinstance(item, int):
            return getattr(self, self.keys(item))
        else:
            raise KeyError(f'Unsupported index type: {type(item)}!')

    def __setitem__(self, key: Union[int, str], value: Any):
        if isinstance(key, str):
            delattr(self, "tensor_attribs")
            return setattr(self, key, value)
        elif isinstance(key, int):
            delattr(self, "tensor_attribs")
            return getattr(self, self.keys(key), value)
        else:
            raise KeyError(f'Unsupported index type: {type(key)}!')

    def __eq__(self, other: Any):
        if isinstance(other, type(self)):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __add__(self, other: "ModelOutput"):
        if self.is_empty:
            return other
        new_inst = copy.deepcopy(self)
        for attrib in self.array_attribs:
            nts = torch.cat((getattr(new_inst, attrib), getattr(other, attrib)), dim=0)
            setattr(new_inst, attrib, nts)
        return new_inst

    def __iadd__(self, other: "ModelOutput"):
        if self.is_empty:
            return other
        for attrib in self.array_attribs:
            nts = torch.cat((getattr(self, attrib), getattr(other, attrib)), dim=0)
            setattr(self, attrib, nts)
        return self



import os
import json
import copy
import logging
import functools
import numpy as np
from typing import List, Optional, Union, Tuple
from seqeval.metrics import classification_report
from seqeval.scheme import IOB2

import torch
from torch.utils.data import DataLoader

from .config import CHMMBaseConfig
from ..data import entity_to_bio_labels, one_hot, probs_to_lbs
from ..embs import build_bert_token_embeddings
from ..base_model.dataset import load_data_from_json, load_data_from_pt

logger = logging.getLogger(__name__)


# noinspection PyBroadException
class CHMMBaseDataset(torch.utils.data.Dataset):
    def __init__(self,
                 text: Optional[List[List[str]]] = None,
                 embs: Optional[List[torch.Tensor]] = None,
                 obs: Optional[List[torch.Tensor]] = None,  # batch, src, token
                 lbs: Optional[List[List[str]]] = None,
                 src: Optional[List[str]] = None,
                 ents: Optional[List[str]] = None
                 ):
        super().__init__()
        self._embs = embs
        self._obs = obs
        self._text = text
        self._lbs = lbs
        self._src = src
        self._ents = ents
        self._src_metrics = None

    @property
    def n_insts(self):
        return len(self.obs)

    @property
    def embs(self):
        return self._embs if self._embs else list()

    @property
    def text(self):
        return self._text if self._text else list()

    @property
    def lbs(self):
        return self._lbs if self._lbs else list()

    @property
    def obs(self):
        return self._obs if self._obs else list()

    @property
    def src(self):
        return self._src

    @property
    def ents(self):
        return self._ents

    @property
    def src_metrics(self):
        if self._src_metrics is None:
            self._src_metrics = self._get_src_metrics()
        return self._src_metrics

    @text.setter
    def text(self, value):
        logger.warning(f'{type(self)}: text has been changed')
        self._text = value

    @obs.setter
    def obs(self, value):
        logger.warning(f'{type(self)}: observations have been changed')
        self._obs = value

    @lbs.setter
    def lbs(self, value):
        logger.warning(f'{type(self)}: labels have been changed')
        self._lbs = value

    @embs.setter
    def embs(self, value):
        logger.warning(f'{type(self)}: embeddings have been changed')
        self._embs = value

    @src.setter
    def src(self, value):
        logger.warning(f'{type(self)}: sources have been changed')
        self._src = value

    @ents.setter
    def ents(self, value):
        logger.warning(f'{type(self)}: entity types have been changed')
        self._ents = value

    @src_metrics.setter
    def src_metrics(self, value):
        self._src_metrics = value

    def __len__(self):
        return self.n_insts

    def __getitem__(self, idx):
        if self._lbs is not None and len(self._lbs) > 0:
            return self._text[idx], self._embs[idx], self._obs[idx], self._lbs[idx]
        else:
            return self._text[idx], self._embs[idx], self._obs[idx]

    def __add__(self, other: "CHMMBaseDataset") -> "CHMMBaseDataset":
        assert self.src and other.src and self.src == other.src, ValueError("Sources not matched!")
        assert self.ents and other.ents and self.ents == other.ents, ValueError("Entity types not matched!")

        return CHMMBaseDataset(
            text=copy.deepcopy(self.text + other.text),
            embs=copy.deepcopy(self.embs + other.embs),
            obs=copy.deepcopy(self.obs + other.obs),
            lbs=copy.deepcopy(self.lbs + other.lbs),
            ents=copy.deepcopy(self.ents),
            src=copy.deepcopy(self.src)
        )

    def __iadd__(self, other: "CHMMBaseDataset") -> "CHMMBaseDataset":
        if self.src:
            assert other.src and self.src == other.src, ValueError("Sources do not match!")
        else:
            assert other.src, ValueError("Attribute `src` not found!")

        if self.ents:
            assert other.ents and self.ents == other.ents, ValueError("Entity types do not match!")
        else:
            assert other.ents, ValueError("Attribute `ents` not found!")

        self.text = copy.deepcopy(self.text + other.text)
        self.embs = copy.deepcopy(self.embs + other.embs)
        self.obs = copy.deepcopy(self.obs + other.obs)
        self.lbs = copy.deepcopy(self.lbs + other.lbs)
        self.ents = copy.deepcopy(other.ents)
        self.src = copy.deepcopy(other.src)
        return self

    def save(self, file_dir: str, dataset_type: str, config: CHMMBaseConfig, force_save: Optional[bool] = False):
        """
        Save dataset for future usage

        Parameters
        ----------
        file_dir: the folder which the dataset will be stored in.
        dataset_type: decides if the dataset is training, validation or test set
        config: configuration file
        force_save: force to save the file even if a file of the same path exists.

        Returns
        -------
        None
        """
        assert dataset_type in ['train', 'valid', 'test']
        output_path = os.path.join(file_dir, f"{dataset_type}.chmmdp")
        if os.path.exists(output_path) and not force_save:
            return None

        chmm_data_dict = {
            'text': self.text,
            'lbs': self.lbs,
            'obs': self.obs,
            'src': self.src,
            'ents': self.ents,
            'embs': self.embs,
            'src_priors': config.src_priors
        }
        torch.save(chmm_data_dict, output_path)
        return None

    def load(self, file_dir: str, dataset_type: str, config: CHMMBaseConfig):
        """
        Load saved datasets and configurations

        Parameters
        ----------
        file_dir: the folder which the dataset is stored in.
        dataset_type: decides if the dataset is training, validation or test set
        config: configuration file

        Returns
        -------
        self
        """
        assert dataset_type in ['train', 'valid', 'test']
        if os.path.isdir(file_dir):
            file_path = os.path.join(file_dir, f'{dataset_type}.chmmdp')
            assert os.path.isfile(file_path), FileNotFoundError(f"{file_path} does not exist!")
        else:
            raise FileNotFoundError(f"{file_dir} does not exist!")

        chmm_data_dict = torch.load(file_path)
        for attr, value in chmm_data_dict.items():
            if attr == 'src_priors':
                continue
            try:
                setattr(self, f'_{attr}', value)
            except AttributeError as err:
                logger.exception(f"Failed to set attribute {attr}: {err}")
                raise err

        config.sources = copy.deepcopy(self.src)
        config.entity_types = copy.deepcopy(self.ents)
        config.bio_label_types = entity_to_bio_labels(self.ents)
        config.src_priors = chmm_data_dict['src_priors']
        config.d_emb = self._embs[0].shape[-1]
        return self

    def load_file(self,
                  file_path: str,
                  config: Optional[CHMMBaseConfig] = None) \
            -> Union["CHMMBaseDataset", Tuple["CHMMBaseDataset", "CHMMBaseConfig"]]:
        """
        Load data from disk

        Parameters
        ----------
        file_path: the directory of the file. In JSON or PT
        config: chmm configuration; Optional to make function testing easier.

        Returns
        -------
        self (MultiSrcNERDataset)
        """
        if config is None:
            has_config_input = False
            config = CHMMBaseConfig()
        else:
            has_config_input = True

        bert_model = getattr(config, "bert_model_name_or_path", 'bert-base_model-uncased')
        device = getattr(config, "device", torch.device('cpu'))

        file_path = os.path.normpath(file_path)
        logger.info(f'Loading data from {file_path}')

        file_dir, file_name = os.path.split(file_path)
        if file_path.endswith('.json'):
            sentence_list, label_list, weak_label_list = load_data_from_json(file_path, config)
            # get embedding directory
            emb_name = f"{'.'.join(file_name.split('.')[:-1])}-emb.pt"
            emb_dir = os.path.join(file_dir, emb_name)
        # for backward compatibility
        elif file_path.endswith('.pt'):
            sentence_list, label_list, weak_label_list = load_data_from_pt(file_path, config)
            # get embedding directory
            emb_dir = file_path.replace('linked', 'emb')
        else:
            logger.error(f"Unsupported data type: {file_path}")
            raise TypeError(f"Unsupported data type: {file_path}")

        if getattr(config, 'load_src_metrics', False):
            try:
                metrics = load_src_metrics(os.path.join(file_dir, 'src_metrics.json'))
                metrics = {k: metrics[k] for k in config.sources}
                self.src_metrics = metrics
            except Exception as e:
                logger.exception(f"Failed to load source metrics: {e}.")

        self._text = sentence_list
        self._lbs = label_list
        self._obs = weak_label_list
        logger.info(f'Data loaded from {file_path}.')

        logger.info(f'Searching for corresponding BERT embeddings...')
        if os.path.isfile(emb_dir):
            logger.info(f"Found embedding file: {emb_dir}. Loading to memory...")
            embs = torch.load(emb_dir)
            if isinstance(embs[0], torch.Tensor):
                self._embs = embs
            elif isinstance(embs[0], np.ndarray):
                self._embs = [torch.from_numpy(emb).to(torch.float) for emb in embs]
            else:
                logger.error(f"Unknown embedding type: {type(embs[0])}")
                raise RuntimeError
        else:
            logger.info(f"{emb_dir} does not exist. Building embeddings instead...")

            if not has_config_input:
                logger.warning(f"No configuration found. Using default bert model `bert-base_model-uncased` "
                               f"and default device `cpu`.")
            self.build_embs(bert_model, device, emb_dir)

        self._src = copy.deepcopy(config.sources)
        self._ents = config.entity_types
        config.d_emb = self._embs[0].shape[-1]
        if getattr(config, 'debug_mode', False):
            self._embs = self._embs[:100]

        # append dummy token/labels in front of the text/lbs/obs
        logger.info("Appending dummy token/labels in front of the text/lbs/obs for CHMM compatibility")
        self._text = [['[CLS]'] + txt for txt in self._text]
        self._lbs = [['O'] + lb for lb in self._lbs]
        prefix = torch.zeros([1, self._obs[0].shape[-2], self._obs[0].shape[-1]])  # shape: 1, n_src, d_obs
        prefix[:, :, 0] = 1
        self._obs = [torch.cat([prefix, inst]) for inst in self._obs]

        if has_config_input:
            return self
        else:
            return self, config

    def build_embs(self,
                   bert_model,
                   device: Optional[torch.device] = torch.device('cpu'),
                   save_dir: Optional[str] = None) -> "CHMMBaseDataset":
        """
        build bert embeddings

        Parameters
        ----------
        bert_model: the location/name of the bert model to use
        device: device
        save_dir: location to update/store the BERT embeddings. Leave None if do not want to save

        Returns
        -------
        self (MultiSrcNERDataset)
        """
        assert bert_model is not None, AssertionError('Please specify BERT model to build embeddings')
        logger.info(f'Building BERT embeddings with {bert_model} on {device}')
        self._embs = build_bert_token_embeddings(self._text, bert_model, bert_model, device, True)
        if save_dir:
            save_dir = os.path.normpath(save_dir)
            logger.info(f'Saving embeddings to {save_dir}...')
            embs = [emb.numpy().astype(np.float32) for emb in self.embs]
            torch.save(embs, save_dir)
        return self

    def update_obs(self,
                   obs: List[List[Union[int, str]]],
                   src_name: str,
                   config: CHMMBaseConfig):
        """
        update weak labels (chmm observations)

        Parameters
        ----------
        obs: input observations (week annotations)
        src_name: source name
        config: configuration file

        Returns
        -------
        self (MultiSrcNERDataset)
        """
        if isinstance(obs[0][0], str):
            lb2ids = {lb: i for i, lb in enumerate(config.bio_label_types)}
            np_map = np.vectorize(lambda lb: lb2ids[lb])
            obs = [np_map(np.asarray(weak_lbs)).tolist() for weak_lbs in obs]

        if len(obs[0]) == len(self.text[0]):
            weak_lbs_one_hot = [one_hot(np.asarray(weak_lbs), n_class=config.n_lbs) for weak_lbs in obs]
        elif len(obs[0]) == len(self.text[0]) - 1:
            weak_lbs_one_hot = [one_hot(np.asarray([0] + weak_lbs), n_class=config.n_lbs) for weak_lbs in obs]
        else:
            logger.error("The length of the input observation does not match the dataset sentences!")
            raise ValueError("The length of the input observation does not match the dataset sentences!")

        if src_name in self._src:
            src_idx = self._src.index(src_name)
            for i in range(len(self._obs)):
                self._obs[i][:, src_idx, :] = torch.tensor(weak_lbs_one_hot[i])
        else:
            self._src.append(src_name)
            for i in range(len(self._obs)):
                self._obs[i] = torch.cat([self._obs[i], torch.tensor(weak_lbs_one_hot[i]).unsqueeze(1)], dim=1)
            # add the source into config and give a heuristic source prior
            if src_name not in config.sources:
                config.sources.append(src_name)
                config.src_priors[src_name] = {ent: (0.7, 0.7) for ent in config.entity_types}

            # remove the cached property
            try:
                delattr(self, "src_metrics")
            except Exception:
                pass

        return self

    def remove_src(self,
                   src_name: str,
                   config: CHMMBaseConfig):
        """
        remove a source and its observations from the dataset

        Parameters
        ----------
        src_name: source name
        config: configuration file

        Returns
        -------
        self (MultiSrcNERDataset)
        """

        if src_name in config.sources:
            config.sources.remove(src_name)
        if src_name in config.src_priors.keys():
            config.src_priors.pop(src_name, None)

        if src_name not in self._src:
            logger.warning(f"Labeling function {src_name} is not presented in dataset. Nothing is changed!")
            return self

        # remove source name
        src_idx = self._src.index(src_name)
        other_idx = list(range(len(self.src)))
        other_idx.remove(src_idx)

        # remove the corresponding observation
        self._src.remove(src_name)
        for i in range(len(self._obs)):
            self._obs[i] = self._obs[i][:, other_idx, :]

        # remove the cached property
        try:
            delattr(self, "src_metrics")
        except Exception:
            pass

        return self

    @functools.cache
    def _get_src_metrics(self):
        src_record_list = [list() for _ in range(len(self.src))]
        for obs in self._obs:
            src_lbs_list = probs_to_lbs(obs, entity_to_bio_labels(self.ents)).T.tolist()
            for src_lbs, src_record in zip(src_lbs_list, src_record_list):
                src_record.append(src_lbs)
        metric_dict = dict()
        for src, record in zip(self.src, src_record_list):
            report = classification_report(
                self.lbs, record, output_dict=True, mode='strict', zero_division=0, scheme=IOB2
            )
            report_dict = dict()
            for ent in self.ents:
                report_dict[ent] = {
                    'precision': report[ent]['precision'],
                    'recall': report[ent]['recall'],
                    'f1': report[ent]['f1-score']
                }
                report_dict['micro-avg'] = {
                    'precision': report['micro avg']['precision'],
                    'recall': report['micro avg']['recall'],
                    'f1': report['micro avg']['f1-score']
                }
            metric_dict[src] = report_dict
        return metric_dict


def batch_prep(emb_list: List[torch.Tensor],
               obs_list: List[torch.Tensor],
               txt_list: Optional[List[List[str]]] = None,
               lbs_list: Optional[List[dict]] = None):
    """
    Pad the instance to the max seq max_seq_length in batch

    All input should already have the dummy element appended to the beginning of the sequence
    """
    for emb, obs, txt, lbs in zip(emb_list, obs_list, txt_list, lbs_list):
        assert len(obs) == len(emb) == len(txt) == len(lbs)
    d_emb = emb_list[0].shape[-1]
    _, n_src, n_obs = obs_list[0].size()
    seq_lens = [len(obs) for obs in obs_list]
    max_seq_len = np.max(seq_lens)

    emb_batch = torch.stack([
        torch.cat([inst, torch.zeros([max_seq_len-len(inst), d_emb])], dim=-2) for inst in emb_list
    ])

    prefix = torch.zeros([1, n_src, n_obs])
    prefix[:, :, 0] = 1
    obs_batch = torch.stack([
        torch.cat([inst, prefix.repeat([max_seq_len-len(inst), 1, 1])]) for inst in obs_list
    ])
    obs_batch /= obs_batch.sum(dim=-1, keepdim=True)

    seq_lens = torch.tensor(seq_lens, dtype=torch.long)

    # we don't need to append the length of txt_list and lbs_list
    return emb_batch, obs_batch, seq_lens, txt_list, lbs_list


def collate_fn(insts):
    """
    Principle used to construct dataloader

    :param insts: original instances
    :return: padded instances
    """
    all_insts = list(zip(*insts))
    if len(all_insts) == 4:
        txt, embs, obs, lbs = all_insts
        batch = batch_prep(emb_list=embs, obs_list=obs, txt_list=txt, lbs_list=lbs)
    elif len(all_insts) == 3:
        txt, embs, obs = all_insts
        batch = batch_prep(emb_list=embs, obs_list=obs, txt_list=txt)
    else:
        logger.error('Unsupported number of instances')
        raise ValueError('Unsupported number of instances')
    return batch


def load_src_metrics(file_path: str):
    """
    Load the file that contains the performance of each source

    Parameters
    ----------
    file_path: file path

    Returns
    -------
    source performance: dict[dict]
    """
    with open(file_path, 'r', encoding='UTF-8') as f:
        metrics = json.load(f)
    return metrics



import logging
from dataclasses import dataclass
from typing import Optional

from ..base_model.config import BaseNERConfig

logger = logging.getLogger(__name__)


@dataclass
class BertBaseConfig(BaseNERConfig):
    """
    Conditional HMM training configuration
    """
    def save(self, file_dir: str, file_name: Optional[str] = 'bert-config') -> "BertBaseConfig":
        BaseNERConfig.save(self, file_dir, file_name)
        return self

    def load(self, file_dir: str, file_name: Optional[str] = 'bert-config') -> "BertBaseConfig":
        BaseNERConfig.load(self, file_dir, file_name)
        return self



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='SeqLbToolkit',
    version='0.1.8',
    author='Yinghao Li',
    author_email='yinghaoli@gatech.edu',
    license='MIT',
    url='https://github.com/Yinghao-Li/seqlbtoolkit',
    description='Commonly-used functions for building sequence labeling models.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='nlp sequence-labeling ml machine-learning natural-language-processing',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Linguistic',
    ],
    packages=find_packages(),
    python_requires=">=3.6",
)



import regex
import itertools
import operator

import numpy as np

from typing import List, Optional, Union
from nltk.tokenize import word_tokenize, sent_tokenize

from .data import merge_list_of_lists


def format_text(text, remove_ref: Optional[bool] = False, remove_emoj: Optional[bool] = False):
    """
    Normalize text and transform some unicode characters into ascii

    Parameters
    ----------
    text: input text string
    remove_ref: whether remove reference tokens such as [1].
    remove_emoj: whether remove emojis from the text

    Returns
    -------
    normalized text string
    """

    # deal with interpuncts
    interpunct = r'[\u00B7\u02D1\u0387\u05BC\u16EB\u2022\u2027\u2218\u2219\u22C5\u23FA' \
                 r'\u25CF\u25E6\u26AB\u2981\u2E30\u2E31\u2E33\u30FB\uA78F\uFF65]'
    text = regex.sub(interpunct, ' ', text)

    # deal with invisible characters
    text = regex.sub(r'[\u2060-\u2061|\u200b]', '', text)

    # deal with bullets
    bullets = r'[\u2022\u2023\u2043\u204C\u204D\u2219\u25CB\u25D8\u25E6' \
              r'\u2619\u2765\u2767\u29BE\u29BF]'
    text = regex.sub(bullets, ' ', text)

    # deal with quotation marks
    s_quotation = r'[\u2018-\u201B\u2039\u203A]'
    d_quotation = r'[\u00AB\u00BB\u201C\u201D\u201E]'
    text = regex.sub(s_quotation, "'", text)
    text = regex.sub(d_quotation, '"', text)

    # deal with overlay tilde
    tilde = r'[\u0303\u223C\u224B\u02DC\u02F7\u223D\u0360\u0334\u0330\u033E' \
            r'\u1DEC\uFE29\uFE2A\uFE22\uFE23]'
    text = regex.sub(tilde, '~', text)

    # deal with overlay not tilde
    not_tilde = r'[\u034A]'
    text = regex.sub(not_tilde, '≁', text)

    text = remove_combining_marks(text)

    # deal with "/"
    text = regex.sub(r'[ ]?/[ ]?', '/', text)

    # deal with invisible Soft Hyphen
    text = regex.sub(r'[\u00ad]', ' ', text)
    # deal with dash/hyphen
    text = regex.sub(r'[\p{Pd}]', '-', text)
    text = regex.sub(r'[\u2212\uf8ff\uf8fe\ue5f8]', '-', text)

    # deal with `=`
    text = regex.sub(r'[﹦＝]', '=', text)
    text = regex.sub(r"([^ ~≈=≅≤≥⩽⩾<>])([~≈=≅≤≥⩽⩾<>]+)([^ ~≈=≅≤≥⩽⩾<>])", r"\g<1> \g<2> \g<3>", text)

    # convert user-defined characters to dash
    text = regex.sub(r'[\uE000-\uF8FF]', '-', text)

    # deal with repeated comma and period
    text = regex.sub(r'[.]+( *[,.])+', r'..', text)

    # deal with white spaces and \n
    text = regex.sub(r'[\p{Z}]', ' ', text)
    text = regex.sub(r'([ \t]+)?[\r\n]([ \t]+)?', ' ', text)
    text = regex.sub(r'\n+', ' ', text)
    text = regex.sub(r'([ ]{2,})', ' ', text)

    if remove_ref:
        text = remove_reference(text)

    if remove_emoj:
        text = remove_emojis(text)

    text = text.strip()
    return text


def remove_emojis(text):
    """
    Remove emojis in the text
    """
    emoj = regex.compile("["
                         u"\U0001F600-\U0001F64F"  # emoticons
                         u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                         u"\U0001F680-\U0001F6FF"  # transport & map symbols
                         u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                         u"\U00002500-\U00002BEF"  # chinese char
                         u"\U00002702-\U000027B0"
                         u"\U00002702-\U000027B0"
                         u"\U000024C2-\U0001F251"
                         u"\U0001f926-\U0001f937"
                         u"\U00010000-\U0010ffff"
                         u"\u2640-\u2642"
                         u"\u2600-\u2B55"
                         u"\u200d"
                         u"\u23cf"
                         u"\u23e9"
                         u"\u231a"
                         u"\ufe0f"  # dingbats
                         u"\u3030]+", regex.UNICODE)
    return regex.sub(emoj, '', text)


def remove_reference(text: str):
    """
    Remove reference tokens such as [1] from text
    """
    # deal with in-sentence references
    text = regex.sub(r' *\[[0-9-, ]+][,-]*', r'', text)
    # deal with after-sentence references
    text = regex.sub(r'([A-Za-zα-ωΑ-Ω\p{Pe}\'"\u2018-\u201d]+[\d-]*)([.])( |[\d]+)([-, ]*[\d]*)*'
                     r'([ ]+[A-Z\dα-ωΑ-Ω\p{Ps}\'"\u2018-\u201d]|$)', r'\g<1>\g<2>\g<5>', text)

    return text


def remove_combining_marks(text: str):
    """
    Remove combining marks (with unicode 0300-036f)

    Parameters
    ----------
    text: input string text

    Returns
    -------
    string text
    """
    # deal with interpuncts
    diacritics = r'[\u0300-\u036F]'
    text = regex.sub(diacritics, '', text)

    return text


def substring_mapping(text: str, mapping_dict: dict):
    """
    Map substrings in the input string according to the dict

    Parameters
    ----------
    text: input string
    mapping_dict: the mapping dictionary

    Returns
    -------
    string text
    """
    rep = dict((regex.escape(k), v) for k, v in mapping_dict.items())
    pattern = regex.compile("|".join(rep.keys()))
    text = pattern.sub(lambda m: rep[regex.escape(m.group(0))], text)
    return text


def split_overlength_bert_input_sequence_legacy(tks: List[str], tokenizer, max_seq_length: Optional[int] = 512):
    """
    Break the sentences that exceeds the maximum BERT length (deprecated)

    Parameters
    ----------
    tks: A list of tokens that are in the original token format (instead of BERT BPE format)
    tokenizer: BERT tokenizer
    max_seq_length: maximum BERT length

    Returns
    -------
    1. sent_tks_list: a list of separated text
    2. the lengths of the broken text
    3. a list of the indices of the separated text
    """

    # Deal with sentences that are longer than 512 BERT tokens
    if len(tokenizer.tokenize(' '.join(tks), add_special_tokens=True)) >= max_seq_length:
        sent_tks_list = [tks]
        bert_length_list = [len(tokenizer.tokenize(' '.join(t), add_special_tokens=True)) for t in sent_tks_list]

        while (np.asarray(bert_length_list) >= max_seq_length).any():
            sep_sent_tks_list = list()

            for tks_list, bert_len in zip(sent_tks_list, bert_length_list):

                if bert_len < max_seq_length:
                    sep_sent_tks_list.append(tks_list)
                    continue

                sep_sent_list = sent_tokenize(' '.join(tks_list))

                sent_lens = list()
                for sep_sent in sep_sent_list:
                    sent_lens.append(len(word_tokenize(sep_sent)))
                end_ids = [np.sum(sent_lens[:i]) for i in range(1, len(sent_lens) + 1)]

                # try to separate sentences as evenly as possible
                halfway_idx = np.argmin((np.array(end_ids) - len(tks_list) / 2) ** 2)
                sep_sent_tks_list.append(tks_list[:end_ids[halfway_idx]])  # split 1
                sep_sent_tks_list.append(tks_list[end_ids[halfway_idx]:])  # split 2

            sent_tks_list = sep_sent_tks_list
            bert_length_list = [len(tokenizer.tokenize(' '.join(t), add_special_tokens=True)) for t in sent_tks_list]

        sent_lengths = [len(s) for s in sent_tks_list]
        assert np.sum(sent_lengths) == len(tks), \
            ValueError(f'Text splitting failed: {tks} ---> {sent_tks_list}')

        return sent_tks_list, sent_lengths, np.arange(len(sent_tks_list))

    else:
        return [tks], [len(tks)], np.array([0], dtype=int)


def split_overlength_bert_input_sequence(sequence: Union[str, List[str], List[List[str]]],
                                         tokenizer,
                                         max_seq_length: Optional[int] = 512) -> List[List[str]]:
    """
    Break the sentences that exceeds the maximum BERT length

    Parameters
    ----------
    sequence: The original text sequence, could be a string, a list tokens,
        or a list of sentences split into tokens (list of lists of string, preferred).
        If the input sequence is in the first two format, it will be split into sentences and tokens using
        `nltk.sent_tokenizer` and `nltk.word_tokenizer`.
    tokenizer: BERT tokenizer used to check input length shen encoded using BERT's vocabulary
    max_seq_length: maximum BERT sequence length

    Returns
    -------
    a list of token collections (list containing tokens from multiple sentences)
    """

    if isinstance(sequence, str):
        tks_seq_list = [word_tokenize(sent) for sent in sent_tokenize(sequence)]
    elif isinstance(sequence[0], str):
        tks_seq_list = [sent.split(' ') for sent in sent_tokenize(' '.join(sequence))]
    elif isinstance(sequence[0][0], str):
        tks_seq_list = sequence
    else:
        raise TypeError("Input parameter `sequence` has Unknown type.")

    tks = merge_list_of_lists(tks_seq_list)
    if len(tokenizer.tokenize(' '.join(tks), add_special_tokens=True)) < max_seq_length:
        return [tks]

    seq_bert_len_list = [len(tokenizer.tokenize(' '.join(tks_seq), add_special_tokens=True))
                         for tks_seq in tks_seq_list]

    if (np.asarray(seq_bert_len_list) > max_seq_length).any():
        raise ValueError("One or more sentences in the input sequence are longer than the designated maximum length.")

    split_points = [0, len(tks_seq_list)]
    split_bert_lens = [sum(seq_bert_len_list[split_points[i]:split_points[i+1]])
                       for i in range(len(split_points)-1)]

    while (np.asarray(split_bert_lens) > max_seq_length).any():

        new_split_points = list()
        for idx, bert_len in enumerate(split_bert_lens):
            if bert_len > max_seq_length:
                seq_bert_len_sub_list = seq_bert_len_list[split_points[idx]:split_points[idx+1]]
                seq_bert_len_sub_accu_list = list(itertools.accumulate(seq_bert_len_sub_list, operator.add))
                # try to separate sentences as evenly as possible
                split_offset = np.argmin((np.array(seq_bert_len_sub_accu_list) - bert_len / 2) ** 2)
                new_split_points.append(split_offset + split_points[idx] + 1)

        split_points += new_split_points
        split_points.sort()

        split_bert_lens = [sum(seq_bert_len_list[split_points[i]:split_points[i+1]])
                           for i in range(len(split_points)-1)]

    split_tks_seq_list = [merge_list_of_lists(tks_seq_list[split_points[i]:split_points[i+1]])
                          for i in range(len(split_points)-1)]

    return split_tks_seq_list


def remove_invalid_parenthesis(sent: str) -> str:
    remove_ind = set()
    stack_ind = []
    for ind, char in enumerate(sent):
        if char not in "()":
            continue
        if char == "(":
            stack_ind.append(ind)
        elif not stack_ind:
            remove_ind.add(ind)
        else:
            stack_ind.pop()
    remove_ind = remove_ind.union(set(stack_ind))
    output = ""
    for ind, char in enumerate(sent):
        if ind in remove_ind:
            continue
        output += char
    return output



import os
import json
import logging
from typing import Optional, List
from dataclasses import dataclass, asdict

logger = logging.getLogger(__name__)


@dataclass
class BaseConfig:
    """
    model & trainer base_model configuration
    """

    def from_args(self, args):
        """
        Initialize configuration from arguments

        Parameters
        ----------
        args: arguments (parent class)

        Returns
        -------
        self (type: BertConfig)
        """
        logger.info(f'Setting {type(self)} from {type(args)}.')
        arg_elements = {attr: getattr(args, attr) for attr in dir(args) if not callable(getattr(args, attr))
                        and not attr.startswith("__") and not attr.startswith("_")}
        logger.info(f'The following attributes will be changed: {arg_elements.keys()}')
        for attr, value in arg_elements.items():
            try:
                setattr(self, attr, value)
            except AttributeError:
                pass
        return self

    def from_config(self, config):
        """
        update configuration results from other config

        Parameters
        ----------
        config: other configurations

        Returns
        -------
        self (BertConfig)
        """
        logger.info(f'Setting {type(self)} from {type(config)}.')
        config_elements = {attr: getattr(config, attr) for attr in dir(self) if not callable(getattr(config, attr))
                           and not attr.startswith("__") and not attr.startswith("_")
                           and getattr(self, attr) is None}
        logger.info(f'The following attributes will be changed: {config_elements.keys()}')
        for attr, value in config_elements.items():
            try:
                setattr(self, attr, value)
            except AttributeError:
                pass
        return self

    def save(self, file_dir: str, file_name: Optional[str] = 'config'):
        """
        Save configuration to file

        Parameters
        ----------
        file_dir: file directory
        file_name: file name (suffix free)

        Returns
        -------
        self
        """
        if os.path.isdir(file_dir):
            file_path = os.path.join(file_dir, f'{file_name}.json')
        elif os.path.isdir(os.path.split(file_dir)[0]):
            file_path = file_dir
        else:
            raise FileNotFoundError(f"{file_dir} does not exist!")

        try:
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(asdict(self), f, ensure_ascii=False, indent=2)
        except Exception as e:
            logger.exception(f"Cannot save config file to {file_path}; "
                             f"encountered Error {e}")
            raise e
        return self

    def load(self, file_dir: str, file_name: Optional[str] = 'config'):
        """
        Load configuration from stored file

        Parameters
        ----------
        file_dir: file directory
        file_name: file name (suffix free)

        Returns
        -------
        self
        """
        if os.path.isdir(file_dir):
            file_path = os.path.join(file_dir, f'{file_name}.json')
            assert os.path.isfile(file_path), FileNotFoundError(f"{file_path} does not exist!")
        elif os.path.isfile(file_dir):
            file_path = file_dir
        else:
            raise FileNotFoundError(f"{file_dir} does not exist!")

        logger.info(f'Setting {type(self)} parameters from {file_path}.')

        with open(file_path, 'r', encoding='utf-8') as f:
            config = json.load(f)
        for attr, value in config.items():
            try:
                setattr(self, attr, value)
            except AttributeError:
                pass
        return self


@dataclass
class BaseNERConfig(BaseConfig):

    entity_types: Optional[List[str]] = None
    bio_label_types: Optional[List[str]] = None

    @property
    def n_lbs(self) -> "int":
        """
        Returns
        -------
        The number of BIO labels
        """
        return len(self.bio_label_types) if self.bio_label_types is not None else 0

    @property
    def n_ent(self) -> "int":
        """
        Returns
        -------
        The number of entity types
        """
        return len(self.entity_types) if self.entity_types is not None else 0

    @property
    def lb2idx(self):
        if self.bio_label_types is not None:
            return {lb: idx for idx, lb in enumerate(self.bio_label_types)}
        else:
            return {}



from seqeval import metrics
from seqeval.scheme import IOB2
from typing import Optional, Union, Any

from ..utils import all_equal


class Metric:
    """
    Sequence labeling metrics

    This class is designed to facilitate easy inheritance.
    You can conveniently add new metrics into the class by reloading the `__init__` function.
    Other member functions should work fine with new metrics.

    Notice that one metric value should not be a `list` instance, otherwise the `append` function
    will not work as intended.
    """
    def __init__(self, precision=None, recall=None, f1=None):
        self.precision = precision
        self.recall = recall
        self.f1 = f1

    def remove_attrs(self, item: Optional[str] = '<default>'):
        """
        Remove one or all default attributes (precision, recall, f1).
        Call this function in your `__init__` if you do not want those metrics

        Returns
        -------
        self
        """
        if item == '<default>':
            delattr(self, 'precision')
            delattr(self, 'recall')
            delattr(self, 'f1')
        else:
            delattr(self, item)
        return self

    def keys(self, idx: Optional[int] = None):
        if idx is None:
            return list(self.__dict__.keys())
        elif isinstance(idx, int):
            return list(self.__dict__.keys())[idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    def items(self, idx: Optional[int] = None):
        if idx is None:
            for k in self.keys():
                yield k, self[k]
        elif isinstance(idx, int):
            for k in self.keys():
                yield k, self[k][idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    def values(self, idx: Optional[int] = None):
        if idx is None:
            for k in self.keys():
                yield self[k]
        elif isinstance(idx, int):
            for k in self.keys():
                yield self[k][idx]
        else:
            raise TypeError(f'Unsupported index type: {type(idx)}!')

    def __len__(self):
        values = self[0]
        if values is None:
            return 0
        elif isinstance(values, (int, float)):
            return 1
        else:
            return len(values)

    def _value_lens(self):
        try:
            return [len(v) for v in self.values()]
        except Exception:
            return None

    def check_equal_length(self):
        lens = self._value_lens()
        if lens is None:
            return False
        return all_equal(lens)

    def __getitem__(self, item: Union[int, str]):
        if isinstance(item, str):
            return getattr(self, item)
        elif isinstance(item, int):
            return getattr(self, self.keys(item))
        else:
            raise KeyError(f'Unsupported index type: {type(item)}!')

    def __setitem__(self, key: Union[int, str], value: Any):
        if isinstance(key, str):
            return setattr(self, key, value)
        elif isinstance(key, int):
            return getattr(self, self.keys(key), value)
        else:
            raise KeyError(f'Unsupported index type: {type(key)}!')

    def __eq__(self, other: Any):
        if isinstance(other, type(self)):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def append(self, m):
        if m is None:
            return self
        self_attr_set = set(self.keys())
        m_attr_set = set(m.keys())
        if m_attr_set - self_attr_set:
            raise ValueError(f'Input instance has unknown attributes: {list(m_attr_set-self_attr_set)}')
        for k in m_attr_set:
            if m[k] is None:
                continue
            if self[k] is None:
                setattr(self, k, m[k])
            else:
                if not isinstance(self[k], list):
                    setattr(self, k, [self[k]])
                if isinstance(m[k], list):
                    self[k] += m[k]
                else:
                    self[k].append(m[k])
        return self

    def pop_attr(self, k: str):
        """
        Pop a given attribute

        Parameters
        ----------
        k: attribute name, string

        Returns
        -------
        popped attribute
        """
        attr = getattr(self, k)
        setattr(self, k, None)
        return attr


# noinspection PyTypeChecker
def get_ner_metrics(true_lbs,
                    pred_lbs,
                    mode: Optional[str] = 'strict',
                    scheme: Optional = IOB2):
    """
    Get NER metrics including precision, recall and f1

    Parameters
    ----------
    true_lbs: true labels
    pred_lbs: predicted labels
    mode:
    scheme: NER label scheme (IOB-2 as default, [O, B-, I-, B-] )

    Returns
    -------
    Metrics
    """
    p = metrics.precision_score(true_lbs, pred_lbs, mode=mode, zero_division=0, scheme=scheme)
    r = metrics.recall_score(true_lbs, pred_lbs, mode=mode, zero_division=0, scheme=scheme)
    f = metrics.f1_score(true_lbs, pred_lbs, mode=mode, zero_division=0, scheme=scheme)
    return Metric(p, r, f)
